import logging
import os

import pytest

import salt.serializers.json as jsonserializer
import salt.serializers.msgpack as msgpackserializer
import salt.serializers.plist as plistserializer
import salt.serializers.python as pythonserializer
import salt.serializers.yaml as yamlserializer
import salt.states.file as filestate
import salt.utils.files
import salt.utils.json
import salt.utils.platform
import salt.utils.win_functions
import salt.utils.yaml
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():
    return {
        filestate: {
            "__env__": "base",
            "__salt__": {"file.manage_file": False},
            "__serializers__": {
                "yaml.serialize": yamlserializer.serialize,
                "yaml.seserialize": yamlserializer.serialize,
                "python.serialize": pythonserializer.serialize,
                "json.serialize": jsonserializer.serialize,
                "plist.serialize": plistserializer.serialize,
                "msgpack.serialize": msgpackserializer.serialize,
            },
            "__opts__": {"test": False, "cachedir": ""},
            "__instance_id__": "",
            "__low__": {},
            "__utils__": {},
        }
    }


# 'directory' function tests: 1
def test_directory():
    """
    Test to ensure that a named directory is present and has the right perms
    """
    name = "/etc/testdir"
    user = "salt"
    group = "saltstack"
    if salt.utils.platform.is_windows():
        name = name.replace("/", "\\")

    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    check_perms_ret = {"name": name, "result": False, "comment": "", "changes": {}}

    comt = "Must provide name to file.directory"
    ret.update({"comment": comt, "name": ""})
    assert filestate.directory("") == ret

    comt = "Cannot specify both max_depth and clean"
    ret.update({"comment": comt, "name": name})
    assert filestate.directory(name, clean=True, max_depth=2) == ret

    mock_t = MagicMock(return_value=True)
    mock_f = MagicMock(return_value=False)
    if salt.utils.platform.is_windows():
        mock_perms = MagicMock(return_value=check_perms_ret)
    else:
        mock_perms = MagicMock(return_value=(check_perms_ret, ""))
    mock_uid = MagicMock(
        side_effect=[
            "",
            "U12",
            "U12",
            "U12",
            "U12",
            "U12",
            "U12",
            "U12",
            "U12",
            "U12",
            "U12",
        ]
    )
    mock_gid = MagicMock(
        side_effect=[
            "",
            "G12",
            "G12",
            "G12",
            "G12",
            "G12",
            "G12",
            "G12",
            "G12",
            "G12",
            "G12",
        ]
    )
    mock_check = MagicMock(
        return_value=(
            None,
            f'The directory "{name}" will be changed',
            {name: {"directory": "new"}},
        )
    )
    mock_error = CommandExecutionError
    with patch.dict(
        filestate.__salt__,
        {
            "config.manage_mode": mock_t,
            "file.user_to_uid": mock_uid,
            "file.group_to_gid": mock_gid,
            "file.stats": mock_f,
            "file.check_perms": mock_perms,
            "file.mkdir": mock_t,
        },
    ), patch("salt.utils.win_dacl.get_sid", mock_error), patch(
        "os.path.isdir", mock_t
    ), patch(
        "salt.states.file._check_directory_win", mock_check
    ):
        if salt.utils.platform.is_windows():
            comt = ""
        else:
            comt = "User salt is not available Group saltstack is not available"
        ret.update({"comment": comt, "name": name})
        assert filestate.directory(name, user=user, group=group) == ret

        with patch.object(os.path, "isabs", mock_f):
            comt = f"Specified file {name} is not an absolute path"
            ret.update({"comment": comt})
            assert filestate.directory(name, user=user, group=group) == ret

        with patch.object(os.path, "isabs", mock_t):
            with patch.object(
                os.path,
                "isfile",
                MagicMock(side_effect=[True, True, False, True, True, True, False]),
            ):
                with patch.object(os.path, "lexists", mock_t):
                    comt = "File exists where the backup target A should go"
                    ret.update({"comment": comt})
                    assert (
                        filestate.directory(
                            name, user=user, group=group, backupname="A"
                        )
                        == ret
                    )

                with patch.object(os.path, "isfile", mock_t):
                    comt = f"Specified location {name} exists and is a file"
                    ret.update({"comment": comt})
                    assert filestate.directory(name, user=user, group=group) == ret

                with patch.object(os.path, "islink", mock_t):
                    comt = f"Specified location {name} exists and is a symlink"
                    ret.update({"comment": comt})
                    assert filestate.directory(name, user=user, group=group) == ret

            with patch.object(os.path, "isdir", mock_f):
                with patch.dict(filestate.__opts__, {"test": True}):
                    if salt.utils.platform.is_windows():
                        comt = 'The directory "{}" will be changed' "".format(name)
                    else:
                        comt = (
                            "The following files will be changed:\n{}:"
                            " directory - new\n".format(name)
                        )
                    ret.update(
                        {
                            "comment": comt,
                            "result": None,
                            "changes": {name: {"directory": "new"}},
                        }
                    )
                    assert filestate.directory(name, user=user, group=group) == ret

                with patch.dict(filestate.__opts__, {"test": False}):
                    with patch.object(os.path, "isdir", mock_f):
                        comt = f"No directory to create {name} in"
                        ret.update({"comment": comt, "result": False})
                        assert filestate.directory(name, user=user, group=group) == ret

                    if salt.utils.platform.is_windows():
                        isdir_side_effect = [False, True, False]
                    else:
                        isdir_side_effect = [True, False, True, False]
                    with patch.object(
                        os.path, "isdir", MagicMock(side_effect=isdir_side_effect)
                    ):
                        comt = f"Failed to create directory {name}"
                        ret.update(
                            {
                                "comment": comt,
                                "result": False,
                                "changes": {name: {"directory": "new"}},
                            }
                        )
                        assert filestate.directory(name, user=user, group=group) == ret

                    check_perms_ret = {
                        "name": name,
                        "result": False,
                        "comment": "",
                        "changes": {},
                    }
                    if salt.utils.platform.is_windows():
                        mock_perms = MagicMock(return_value=check_perms_ret)
                    else:
                        mock_perms = MagicMock(return_value=(check_perms_ret, ""))

                    recurse = ["silent"]
                    ret = {
                        "name": name,
                        "result": False,
                        "comment": "Directory /etc/testdir updated",
                        "changes": {"recursion": "Changes silenced"},
                    }
                    if salt.utils.platform.is_windows():
                        ret["comment"] = ret["comment"].replace("/", "\\")
                    with patch.dict(
                        filestate.__salt__, {"file.check_perms": mock_perms}
                    ):
                        with patch.object(os.path, "isdir", mock_t):
                            assert (
                                filestate.directory(
                                    name, user=user, recurse=recurse, group=group
                                )
                                == ret
                            )

                    check_perms_ret = {
                        "name": name,
                        "result": False,
                        "comment": "",
                        "changes": {},
                    }
                    if salt.utils.platform.is_windows():
                        mock_perms = MagicMock(return_value=check_perms_ret)
                    else:
                        mock_perms = MagicMock(return_value=(check_perms_ret, ""))

                    recurse = ["ignore_files", "ignore_dirs"]
                    ret = {
                        "name": name,
                        "result": False,
                        "comment": 'Must not specify "recurse" '
                        'options "ignore_files" and '
                        '"ignore_dirs" at the same '
                        "time.",
                        "changes": {},
                    }
                    with patch.dict(
                        filestate.__salt__, {"file.check_perms": mock_perms}
                    ):
                        with patch.object(os.path, "isdir", mock_t):
                            assert (
                                filestate.directory(
                                    name, user=user, recurse=recurse, group=group
                                )
                                == ret
                            )

                    comt = f"Directory {name} updated"
                    ret = {
                        "name": name,
                        "result": True,
                        "comment": comt,
                        "changes": {"group": "group", "mode": "0777", "user": "user"},
                    }

                    check_perms_ret = {
                        "name": name,
                        "result": True,
                        "comment": "",
                        "changes": {"group": "group", "mode": "0777", "user": "user"},
                    }

                    if salt.utils.platform.is_windows():
                        _mock_perms = MagicMock(return_value=check_perms_ret)
                    else:
                        _mock_perms = MagicMock(return_value=(check_perms_ret, ""))
                    with patch.object(os.path, "isdir", mock_t):
                        with patch.dict(
                            filestate.__salt__, {"file.check_perms": _mock_perms}
                        ):
                            assert (
                                filestate.directory(name, user=user, group=group) == ret
                            )


def test_directory_test_mode_user_group_not_present():
    name = "/etc/testdir"
    user = "salt"
    group = "saltstack"
    if salt.utils.platform.is_windows():
        name = name.replace("/", "\\")

    ret = {
        "name": name,
        "result": None,
        "comment": "",
        "changes": {name: {"directory": "new"}},
    }

    if salt.utils.platform.is_windows():
        comt = 'The directory "{}" will be changed' "".format(name)
    else:
        comt = "The following files will be changed:\n{}:" " directory - new\n".format(
            name
        )
    ret["comment"] = comt

    mock_f = MagicMock(return_value=False)
    mock_uid = MagicMock(
        side_effect=[
            "",
            "U12",
            "",
        ]
    )
    mock_gid = MagicMock(
        side_effect=[
            "G12",
            "",
            "",
        ]
    )
    mock_error = CommandExecutionError
    with patch.dict(
        filestate.__salt__,
        {
            "file.user_to_uid": mock_uid,
            "file.group_to_gid": mock_gid,
            "file.stats": mock_f,
        },
    ), patch("salt.utils.win_dacl.get_sid", mock_error), patch.object(
        os.path, "isdir", mock_f
    ), patch.dict(
        filestate.__opts__, {"test": True}
    ):
        assert filestate.directory(name, user=user, group=group) == ret
        assert filestate.directory(name, user=user, group=group) == ret
        assert filestate.directory(name, user=user, group=group) == ret



import pytest

import salt.exceptions
import salt.returners.mongo_future_return as mongo
from salt.utils.versions import Version
from tests.support.mock import patch


@pytest.fixture
def configure_loader_modules():

    return {
        mongo: {
            "__opts__": {
                "mongo.uri": "mongodb://root:pass@localhost27017/salt?authSource=admin"
            }
        }
    }


def test_config_exception():
    opts = {
        "mongo.host": "localhost",
        "mongo.port": 27017,
        "mongo.user": "root",
        "mongo.password": "pass",
        "mongo.uri": "mongodb://root:pass@localhost27017/salt?authSource=admin",
    }
    with patch(
        "salt.returners.mongo_future_return.PYMONGO_VERSION",
        Version("4.3.2"),
        create=True,
    ), patch.dict(mongo.__opts__, opts):
        with pytest.raises(salt.exceptions.SaltConfigurationError):
            mongo.returner({})



"""
Minion data cache plugin for MySQL database.

.. versionadded:: 2018.3.0

It is up to the system administrator to set up and configure the MySQL
infrastructure. All is needed for this plugin is a working MySQL server.

.. warning::

    The mysql.database and mysql.table_name will be directly added into certain
    queries. Salt treats these as trusted input.

The module requires the database (default ``salt_cache``) to exist but creates
its own table if needed. The keys are indexed using the ``bank`` and
``etcd_key`` columns.

To enable this cache plugin, the master will need the python client for
MySQL installed. This can be easily installed with pip:

.. code-block:: bash

    pip install pymysql

Optionally, depending on the MySQL agent configuration, the following values
could be set in the master config. These are the defaults:

.. code-block:: yaml

    mysql.host: 127.0.0.1
    mysql.port: 2379
    mysql.user: None
    mysql.password: None
    mysql.database: salt_cache
    mysql.table_name: cache

Related docs can be found in the `python-mysql documentation`_.

To use the mysql as a minion data cache backend, set the master ``cache`` config
value to ``mysql``:

.. code-block:: yaml

    cache: mysql


.. _`MySQL documentation`: https://github.com/coreos/mysql
.. _`python-mysql documentation`: http://python-mysql.readthedocs.io/en/latest/

"""

import copy
import logging
import time

import salt.payload
import salt.utils.stringutils
from salt.exceptions import SaltCacheError

try:
    # Trying to import MySQLdb
    import MySQLdb
    import MySQLdb.converters
    import MySQLdb.cursors
    from MySQLdb.connections import OperationalError
except ImportError:
    try:
        # MySQLdb import failed, try to import PyMySQL
        import pymysql

        pymysql.install_as_MySQLdb()
        import MySQLdb
        import MySQLdb.converters
        import MySQLdb.cursors
        from MySQLdb.err import OperationalError
    except ImportError:
        MySQLdb = None


_DEFAULT_DATABASE_NAME = "salt_cache"
_DEFAULT_CACHE_TABLE_NAME = "cache"
_RECONNECT_INTERVAL_SEC = 0.050

log = logging.getLogger(__name__)

# Module properties

__virtualname__ = "mysql"
__func_alias__ = {"ls": "list"}


def __virtual__():
    """
    Confirm that a python mysql client is installed.
    """
    return bool(MySQLdb), "No python mysql client installed." if MySQLdb is None else ""


def force_reconnect():
    """
    Force a reconnection to the MySQL database, by removing the client from
    Salt's __context__.
    """
    __context__.pop("mysql_client", None)


def run_query(conn, query, args=None, retries=3):
    """
    Get a cursor and run a query. Reconnect up to ``retries`` times if
    needed.
    Returns: cursor, affected rows counter
    Raises: SaltCacheError, AttributeError, OperationalError
    """
    if conn is None:
        conn = __context__.get("mysql_client")
    try:
        cur = conn.cursor()

        if not args:
            log.debug("Doing query: %s", query)
            out = cur.execute(query)
        else:
            log.debug("Doing query: %s args: %s ", query, repr(args))
            out = cur.execute(query, args)

        return cur, out
    except (AttributeError, OperationalError) as e:
        if retries == 0:
            raise
        # reconnect creating new client
        time.sleep(_RECONNECT_INTERVAL_SEC)
        if conn is None:
            log.debug("mysql_cache: creating db connection")
        else:
            log.info("mysql_cache: recreating db connection due to: %r", e)
        __context__["mysql_client"] = MySQLdb.connect(**__context__["mysql_kwargs"])
        return run_query(
            conn=__context__.get("mysql_client"),
            query=query,
            args=args,
            retries=(retries - 1),
        )
    except Exception as e:  # pylint: disable=broad-except
        if len(query) > 150:
            query = query[:150] + "<...>"
        raise SaltCacheError(
            "Error running {}{}: {}".format(query, f"- args: {args}" if args else "", e)
        )


def _create_table():
    """
    Create table if needed
    """
    # Explicitly check if the table already exists as the library logs a
    # warning on CREATE TABLE
    query = """SELECT COUNT(TABLE_NAME) FROM information_schema.tables
        WHERE table_schema = %s AND table_name = %s"""
    cur, _ = run_query(
        __context__.get("mysql_client"),
        query,
        args=(__context__["mysql_kwargs"]["db"], __context__["mysql_table_name"]),
    )
    r = cur.fetchone()
    cur.close()
    if r[0] == 1:
        query = """
        SELECT COUNT(TABLE_NAME)
        FROM
            information_schema.columns
        WHERE
            table_schema = %s
            AND table_name = %s
            AND column_name = 'last_update'
        """
        cur, _ = run_query(
            __context__["mysql_client"],
            query,
            args=(__context__["mysql_kwargs"]["db"], __context__["mysql_table_name"]),
        )
        r = cur.fetchone()
        cur.close()
        if r[0] == 1:
            return
        else:
            query = """
            ALTER TABLE {}.{}
            ADD COLUMN last_update TIMESTAMP NOT NULL
                                   DEFAULT CURRENT_TIMESTAMP
                                   ON UPDATE CURRENT_TIMESTAMP
            """.format(
                __context__["mysql_kwargs"]["db"], __context__["mysql_table_name"]
            )
            cur, _ = run_query(__context__["mysql_client"], query)
            cur.close()
            return

    query = """CREATE TABLE IF NOT EXISTS {} (
      bank CHAR(255),
      etcd_key CHAR(255),
      data MEDIUMBLOB,
      last_update TIMESTAMP NOT NULL
                  DEFAULT CURRENT_TIMESTAMP
                  ON UPDATE CURRENT_TIMESTAMP,
      PRIMARY KEY(bank, etcd_key)
    );""".format(
        __context__["mysql_table_name"]
    )
    log.info("mysql_cache: creating table %s", __context__["mysql_table_name"])
    cur, _ = run_query(__context__.get("mysql_client"), query)
    cur.close()


def _init_client():
    """Initialize connection and create table if needed"""
    if __context__.get("mysql_client") is not None:
        return

    opts = copy.deepcopy(__opts__)
    mysql_kwargs = {
        "autocommit": True,
        "host": opts.pop("mysql.host", "127.0.0.1"),
        "user": opts.pop("mysql.user", None),
        "passwd": opts.pop("mysql.password", None),
        "db": opts.pop("mysql.database", _DEFAULT_DATABASE_NAME),
        "port": opts.pop("mysql.port", 3306),
        "unix_socket": opts.pop("mysql.unix_socket", None),
        "connect_timeout": opts.pop("mysql.connect_timeout", None),
    }
    mysql_kwargs["autocommit"] = True

    __context__["mysql_table_name"] = opts.pop("mysql.table_name", "salt")

    # Gather up any additional MySQL configuration options
    for k in opts:
        if k.startswith("mysql."):
            _key = k.split(".")[1]
            mysql_kwargs[_key] = opts.get(k)

    # TODO: handle SSL connection parameters

    for k, v in copy.deepcopy(mysql_kwargs).items():
        if v is None:
            mysql_kwargs.pop(k)
    kwargs_copy = mysql_kwargs.copy()
    kwargs_copy["passwd"] = "<hidden>"
    log.info("mysql_cache: Setting up client with params: %r", kwargs_copy)
    __context__["mysql_kwargs"] = mysql_kwargs
    # The MySQL client is created later on by run_query
    _create_table()


def store(bank, key, data):
    """
    Store a key value.
    """
    _init_client()
    data = salt.payload.dumps(data)
    query = "REPLACE INTO {} (bank, etcd_key, data) values(%s,%s,%s)".format(
        __context__["mysql_table_name"]
    )
    args = (bank, key, data)

    cur, cnt = run_query(__context__.get("mysql_client"), query, args=args)
    cur.close()
    if cnt not in (1, 2):
        raise SaltCacheError(f"Error storing {bank} {key} returned {cnt}")


def fetch(bank, key):
    """
    Fetch a key value.
    """
    _init_client()
    query = "SELECT data FROM {} WHERE bank=%s AND etcd_key=%s".format(
        __context__["mysql_table_name"]
    )
    cur, _ = run_query(__context__.get("mysql_client"), query, args=(bank, key))
    r = cur.fetchone()
    cur.close()
    if r is None:
        return {}
    return salt.payload.loads(r[0])


def flush(bank, key=None):
    """
    Remove the key from the cache bank with all the key content.
    """
    _init_client()
    query = "DELETE FROM {} WHERE bank=%s".format(__context__["mysql_table_name"])
    if key is None:
        data = (bank,)
    else:
        data = (bank, key)
        query += " AND etcd_key=%s"

    cur, _ = run_query(__context__["mysql_client"], query, args=data)
    cur.close()


def ls(bank):
    """
    Return an iterable object containing all entries stored in the specified
    bank.
    """
    _init_client()
    query = "SELECT etcd_key FROM {} WHERE bank=%s".format(
        __context__["mysql_table_name"]
    )
    cur, _ = run_query(__context__.get("mysql_client"), query, args=(bank,))
    out = [row[0] for row in cur.fetchall()]
    cur.close()
    return out


def contains(bank, key):
    """
    Checks if the specified bank contains the specified key.
    """
    _init_client()
    if key is None:
        data = (bank,)
        query = "SELECT COUNT(data) FROM {} WHERE bank=%s".format(
            __context__["mysql_table_name"]
        )
    else:
        data = (bank, key)
        query = "SELECT COUNT(data) FROM {} WHERE bank=%s AND etcd_key=%s".format(
            __context__["mysql_table_name"]
        )
    cur, _ = run_query(__context__.get("mysql_client"), query, args=data)
    r = cur.fetchone()
    cur.close()
    return r[0] == 1


def updated(bank, key):
    """
    Return the integer Unix epoch update timestamp of the specified bank and
    key.
    """
    _init_client()
    query = (
        "SELECT UNIX_TIMESTAMP(last_update) FROM {} WHERE bank=%s "
        "AND etcd_key=%s".format(__context__["mysql_table_name"])
    )
    data = (bank, key)
    cur, _ = run_query(__context__["mysql_client"], query=query, args=data)
    r = cur.fetchone()
    cur.close()
    return int(r[0]) if r else r



#
# Copyright 2015 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Philips HUE lamps module for proxy.

.. versionadded:: 2015.8.3

First create a new user on the Hue bridge by following the
`Meet hue <https://www.developers.meethue.com/documentation/getting-started>`_ instructions.

To configure the proxy minion:

.. code-block:: yaml

    proxy:
      proxytype: philips_hue
      host: [hostname or ip]
      user: [username]

"""

# pylint: disable=import-error,no-name-in-module,redefined-builtin

import http.client
import logging
import time

import salt.utils.json
from salt.exceptions import CommandExecutionError, MinionError

__proxyenabled__ = ["philips_hue"]

CONFIG = {}
log = logging.getLogger(__file__)


class Const:
    """
    Constants for the lamp operations.
    """

    LAMP_ON = {"on": True, "transitiontime": 0}
    LAMP_OFF = {"on": False, "transitiontime": 0}

    COLOR_WHITE = {"xy": [0.3227, 0.329]}
    COLOR_DAYLIGHT = {"xy": [0.3806, 0.3576]}
    COLOR_RED = {"hue": 0, "sat": 254}
    COLOR_GREEN = {"hue": 25500, "sat": 254}
    COLOR_ORANGE = {"hue": 12000, "sat": 254}
    COLOR_PINK = {"xy": [0.3688, 0.2095]}
    COLOR_BLUE = {"hue": 46920, "sat": 254}
    COLOR_YELLOW = {"xy": [0.4432, 0.5154]}
    COLOR_PURPLE = {"xy": [0.3787, 0.1724]}


def __virtual__():
    """
    Validate the module.
    """
    return True


def init(cnf):
    """
    Initialize the module.
    """
    CONFIG["host"] = cnf.get("proxy", {}).get("host")
    if not CONFIG["host"]:
        raise MinionError(
            message="Cannot find 'host' parameter in the proxy configuration"
        )

    CONFIG["user"] = cnf.get("proxy", {}).get("user")
    if not CONFIG["user"]:
        raise MinionError(
            message="Cannot find 'user' parameter in the proxy configuration"
        )

    CONFIG["uri"] = "/api/{}".format(CONFIG["user"])


def ping(*args, **kw):
    """
    Ping the lamps.
    """
    # Here blink them
    return True


def shutdown(opts, *args, **kw):
    """
    Shuts down the service.
    """
    # This is no-op method, which is required but makes nothing at this point.
    return True


def _query(lamp_id, state, action="", method="GET"):
    """
    Query the URI

    :return:
    """
    # Because salt.utils.query is that dreadful... :(

    err = None
    url = "{}/lights{}".format(CONFIG["uri"], lamp_id and f"/{lamp_id}" or "") + (
        action and f"/{action}" or ""
    )
    conn = http.client.HTTPConnection(CONFIG["host"])
    if method == "PUT":
        conn.request(method, url, salt.utils.json.dumps(state))
    else:
        conn.request(method, url)
    resp = conn.getresponse()

    if resp.status == http.client.OK:
        res = salt.utils.json.loads(resp.read())
    else:
        err = f"HTTP error: {resp.status}, {resp.reason}"
    conn.close()
    if err:
        raise CommandExecutionError(err)

    return res


def _set(lamp_id, state, method="state"):
    """
    Set state to the device by ID.

    :param lamp_id:
    :param state:
    :return:
    """
    try:
        res = _query(lamp_id, state, action=method, method="PUT")
    except Exception as err:  # pylint: disable=broad-except
        raise CommandExecutionError(err)

    res = len(res) > 1 and res[-1] or res[0]
    if res.get("success"):
        res = {"result": True}
    elif res.get("error"):
        res = {
            "result": False,
            "description": res["error"]["description"],
            "type": res["error"]["type"],
        }

    return res


def _get_devices(params):
    """
    Parse device(s) ID(s) from the common params.

    :param params:
    :return:
    """
    if "id" not in params:
        raise CommandExecutionError("Parameter ID is required.")

    return (
        isinstance(params["id"], int)
        and [params["id"]]
        or [int(dev) for dev in params["id"].split(",")]
    )


def _get_lights():
    """
    Get all available lighting devices.
    """
    return _query(None, None)


# Callers
def call_lights(*args, **kwargs):
    """
    Get info about all available lamps.

    Options:

    * **id**: Specifies a device ID. Can be a comma-separated values. All, if omitted.

    CLI Example:

    .. code-block:: bash

        salt '*' hue.lights
        salt '*' hue.lights id=1
        salt '*' hue.lights id=1,2,3
    """
    res = dict()
    lights = _get_lights()
    for dev_id in "id" in kwargs and _get_devices(kwargs) or sorted(lights.keys()):
        if lights.get(str(dev_id)):
            res[dev_id] = lights[str(dev_id)]

    return res or False


def call_switch(*args, **kwargs):
    """
    Switch lamp ON/OFF.

    If no particular state is passed,
    then lamp will be switched to the opposite state.

    Options:

    * **id**: Specifies a device ID. Can be a comma-separated values. All, if omitted.
    * **on**: True or False. Inverted current, if omitted

    CLI Example:

    .. code-block:: bash

        salt '*' hue.switch
        salt '*' hue.switch id=1
        salt '*' hue.switch id=1,2,3 on=True
    """
    out = dict()
    devices = _get_lights()
    for dev_id in "id" not in kwargs and sorted(devices.keys()) or _get_devices(kwargs):
        if "on" in kwargs:
            state = kwargs["on"] and Const.LAMP_ON or Const.LAMP_OFF
        else:
            # Invert the current state
            state = (
                devices[str(dev_id)]["state"]["on"] and Const.LAMP_OFF or Const.LAMP_ON
            )
        out[dev_id] = _set(dev_id, state)

    return out


def call_blink(*args, **kwargs):
    """
    Blink a lamp. If lamp is ON, then blink ON-OFF-ON, otherwise OFF-ON-OFF.

    Options:

    * **id**: Specifies a device ID. Can be a comma-separated values. All, if omitted.
    * **pause**: Time in seconds. Can be less than 1, i.e. 0.7, 0.5 sec.

    CLI Example:

    .. code-block:: bash

        salt '*' hue.blink id=1
        salt '*' hue.blink id=1,2,3
    """
    devices = _get_lights()
    pause = kwargs.get("pause", 0)
    res = dict()
    for dev_id in "id" not in kwargs and sorted(devices.keys()) or _get_devices(kwargs):
        state = devices[str(dev_id)]["state"]["on"]
        _set(dev_id, state and Const.LAMP_OFF or Const.LAMP_ON)
        if pause:
            time.sleep(pause)
        res[dev_id] = _set(dev_id, not state and Const.LAMP_OFF or Const.LAMP_ON)

    return res


def call_ping(*args, **kwargs):
    """
    Ping the lamps by issuing a short inversion blink to all available devices.

    CLI Example:

    .. code-block:: bash

        salt '*' hue.ping
    """
    errors = dict()
    for dev_id, dev_status in call_blink().items():
        if not dev_status["result"]:
            errors[dev_id] = False

    return errors or True


def call_status(*args, **kwargs):
    """
    Return the status of the lamps.

    Options:

    * **id**: Specifies a device ID. Can be a comma-separated values. All, if omitted.

    CLI Example:

    .. code-block:: bash

        salt '*' hue.status
        salt '*' hue.status id=1
        salt '*' hue.status id=1,2,3
    """
    res = dict()
    devices = _get_lights()
    for dev_id in "id" not in kwargs and sorted(devices.keys()) or _get_devices(kwargs):
        dev_id = str(dev_id)
        res[dev_id] = {
            "on": devices[dev_id]["state"]["on"],
            "reachable": devices[dev_id]["state"]["reachable"],
        }

    return res


def call_rename(*args, **kwargs):
    """
    Rename a device.

    Options:

    * **id**: Specifies a device ID. Only one device at a time.
    * **title**: Title of the device.

    CLI Example:

    .. code-block:: bash

        salt '*' hue.rename id=1 title='WC for cats'
    """
    dev_id = _get_devices(kwargs)
    if len(dev_id) > 1:
        raise CommandExecutionError("Only one device can be renamed at a time")

    if "title" not in kwargs:
        raise CommandExecutionError("Title is missing")

    return _set(dev_id[0], {"name": kwargs["title"]}, method="")


def call_alert(*args, **kwargs):
    """
    Lamp alert

    Options:

    * **id**: Specifies a device ID. Can be a comma-separated values. All, if omitted.
    * **on**: Turns on or off an alert. Default is True.

    CLI Example:

    .. code-block:: bash

        salt '*' hue.alert
        salt '*' hue.alert id=1
        salt '*' hue.alert id=1,2,3 on=false
    """
    res = dict()

    devices = _get_lights()
    for dev_id in "id" not in kwargs and sorted(devices.keys()) or _get_devices(kwargs):
        res[dev_id] = _set(
            dev_id, {"alert": kwargs.get("on", True) and "lselect" or "none"}
        )

    return res


def call_effect(*args, **kwargs):
    """
    Set an effect to the lamp.

    Options:

    * **id**: Specifies a device ID. Can be a comma-separated values. All, if omitted.
    * **type**: Type of the effect. Possible values are "none" or "colorloop". Default "none".

    CLI Example:

    .. code-block:: bash

        salt '*' hue.effect
        salt '*' hue.effect id=1
        salt '*' hue.effect id=1,2,3 type=colorloop
    """
    res = dict()

    devices = _get_lights()
    for dev_id in "id" not in kwargs and sorted(devices.keys()) or _get_devices(kwargs):
        res[dev_id] = _set(dev_id, {"effect": kwargs.get("type", "none")})

    return res


def call_color(*args, **kwargs):
    """
    Set a color to the lamp.

    Options:

    * **id**: Specifies a device ID. Can be a comma-separated values. All, if omitted.
    * **color**: Fixed color. Values are: red, green, blue, orange, pink, white,
                 yellow, daylight, purple. Default white.
    * **transition**: Transition 0~200.

    Advanced:

    * **gamut**: XY coordinates. Use gamut according to the Philips HUE devices documentation.
                 More: http://www.developers.meethue.com/documentation/hue-xy-values

    CLI Example:

    .. code-block:: bash

        salt '*' hue.color
        salt '*' hue.color id=1
        salt '*' hue.color id=1,2,3 oolor=red transition=30
        salt '*' hue.color id=1 gamut=0.3,0.5
    """
    res = dict()

    colormap = {
        "red": Const.COLOR_RED,
        "green": Const.COLOR_GREEN,
        "blue": Const.COLOR_BLUE,
        "orange": Const.COLOR_ORANGE,
        "pink": Const.COLOR_PINK,
        "white": Const.COLOR_WHITE,
        "yellow": Const.COLOR_YELLOW,
        "daylight": Const.COLOR_DAYLIGHT,
        "purple": Const.COLOR_PURPLE,
    }

    devices = _get_lights()
    color = kwargs.get("gamut")
    if color:
        color = color.split(",")
        if len(color) == 2:
            try:
                color = {"xy": [float(color[0]), float(color[1])]}
            except Exception as ex:  # pylint: disable=broad-except
                color = None
        else:
            color = None

    if not color:
        color = colormap.get(kwargs.get("color", "white"), Const.COLOR_WHITE)
    color.update({"transitiontime": max(min(kwargs.get("transition", 0), 200), 0)})

    for dev_id in "id" not in kwargs and sorted(devices.keys()) or _get_devices(kwargs):
        res[dev_id] = _set(dev_id, color)

    return res


def call_brightness(*args, **kwargs):
    """
    Set an effect to the lamp.

    Arguments:

    * **value**: 0~255 brightness of the lamp.

    Options:

    * **id**: Specifies a device ID. Can be a comma-separated values. All, if omitted.
    * **transition**: Transition 0~200. Default 0.

    CLI Example:

    .. code-block:: bash

        salt '*' hue.brightness value=100
        salt '*' hue.brightness id=1 value=150
        salt '*' hue.brightness id=1,2,3 value=255
    """
    res = dict()

    if "value" not in kwargs:
        raise CommandExecutionError("Parameter 'value' is missing")

    try:
        brightness = max(min(int(kwargs["value"]), 244), 1)
    except Exception as err:  # pylint: disable=broad-except
        raise CommandExecutionError("Parameter 'value' does not contains an integer")

    try:
        transition = max(min(int(kwargs["transition"]), 200), 0)
    except Exception as err:  # pylint: disable=broad-except
        transition = 0

    devices = _get_lights()
    for dev_id in "id" not in kwargs and sorted(devices.keys()) or _get_devices(kwargs):
        res[dev_id] = _set(dev_id, {"bri": brightness, "transitiontime": transition})

    return res


def call_temperature(*args, **kwargs):
    """
    Set the mired color temperature. More: http://en.wikipedia.org/wiki/Mired

    Arguments:

    * **value**: 150~500.

    Options:

    * **id**: Specifies a device ID. Can be a comma-separated values. All, if omitted.

    CLI Example:

    .. code-block:: bash

        salt '*' hue.temperature value=150
        salt '*' hue.temperature value=150 id=1
        salt '*' hue.temperature value=150 id=1,2,3
    """
    res = dict()

    if "value" not in kwargs:
        raise CommandExecutionError("Parameter 'value' (150~500) is missing")
    try:
        value = max(min(int(kwargs["value"]), 500), 150)
    except Exception as err:  # pylint: disable=broad-except
        raise CommandExecutionError("Parameter 'value' does not contains an integer")

    devices = _get_lights()
    for dev_id in "id" not in kwargs and sorted(devices.keys()) or _get_devices(kwargs):
        res[dev_id] = _set(dev_id, {"ct": value})

    return res



"""
Manage Grafana Dashboards

This module uses ``elasticsearch``, which can be installed via package, or pip.

You can specify elasticsearch hosts directly to the module, or you can use an
elasticsearch profile via pillars:

.. code-block:: yaml

    mygrafanaprofile:
      hosts:
        - es1.example.com:9200
        - es2.example.com:9200
      index: grafana-dash

.. code-block:: yaml

    # Basic usage (uses default pillar profile key 'grafana')
    Ensure myservice dashboard is managed:
      grafana.dashboard_present:
        - name: myservice
        - dashboard_from_pillar: default
        - rows_from_pillar:
            - systemhealth
            - requests

    # Passing hosts in
    Ensure myservice dashboard is managed:
      grafana.dashboard_present:
        - name: myservice
        - dashboard_from_pillar: default
        - rows:
            - collapse: false
              editable: true
              height: 150px
              title: System Health
              panels:
                - aliasColors: {}
                  id: 200000
                  annotate:
                    enable: false
                  bars: false
                  datasource: null
                  editable: true
                  error: false
                  fill: 7
                  grid:
                    leftMax: 100
                    leftMin: null
                    rightMax: null
                    rightMin: null
                    threshold1: 60
                    threshold1Color: rgb(216, 27, 27)
                    threshold2: null
                    threshold2Color: rgba(234, 112, 112, 0.22)
                  leftYAxisLabel: ''
                  legend:
                    avg: false
                    current: false
                    max: false
                    min: false
                    show: false
                    total: false
                    values: false
                  lines: true
                  linewidth: 1
                  nullPointMode: connected
                  percentage: false
                  pointradius: 5
                  points: false
                  renderer: flot
                  resolution: 100
                  scale: 1
                  seriesOverrides: []
                  span: 4
                  stack: false
                  steppedLine: false
                  targets:
                    - target: cloudwatch.aws.ec2.mysrv.cpuutilization.average
                  title: CPU (asg average)
                  tooltip:
                    query_as_alias: true
                    shared: false
                    value_type: cumulative
                  type: graph
                  x-axis: true
                  y-axis: true
                  y_formats:
                    - short
                    - short
                  zerofill: true
        - rows_from_pillar:
          - systemhealth
          - requests
        - profile:
            hosts:
              - es1.example.com:9200
              - es2.example.com:9200
            index: grafana-dash

    # Using a profile from pillars
    Ensure myservice dashboard is managed:
      grafana.dashboard_present:
        - name: myservice
        - dashboard:
            annotations:
              enable: true
              list: []
            editable: true
            hideAllLegends: false
            hideControls: false
            nav:
              - collapse: false
                enable: true
                notice: false
                now: true
                refresh_intervals:
                  - 10s
                  - 30s
                  - 1m
                  - 5m
                  - 15m
                  - 30m
                  - 1h
                  - 2h
                  - 1d
                status: Stable
                time_options:
                  - 5m
                  - 15m
                  - 1h
                  - 2h
                  - 3h
                  - 4h
                  - 6h
                  - 12h
                  - 1d
                  - 2d
                  - 4d
                  - 7d
                  - 16d
                  - 30d
                type: timepicker
            originalTitle: dockerregistry
            refresh: 1m
            rows: []
            sharedCrosshair: false
            style: dark
            tags: []
            templating:
              enable: true
              list: []
            time:
              from: now-2h
              to: now
            timezone: browser
        - rows_from_pillars:
          - systemhealth
          - requests
        - profile: mygrafanaprofile

The behavior of this module is to create dashboards if they do not exist, to
add rows if they do not exist in existing dashboards, and to update rows if
they exist in dashboards. The module will not manage rows that are not defined,
allowing users to manage their own custom rows.
"""

import copy

import salt.utils.json
from salt.exceptions import SaltInvocationError
from salt.utils.dictdiffer import DictDiffer


def __virtual__():
    """
    Only load if grafana is available.
    """
    if "elasticsearch.exists" in __salt__:
        return "grafana"
    return (False, "elasticsearch module could not be loaded")


def _parse_profile(profile):
    """
    From a pillar key, or a dictionary, return index and host keys.
    """
    if isinstance(profile, str):
        _profile = __salt__["config.option"](profile)
        if not _profile:
            msg = f"Pillar key for profile {profile} not found."
            raise SaltInvocationError(msg)
    else:
        _profile = profile
    hosts = _profile.get("hosts")
    index = _profile.get("index")
    return (hosts, index)


def _rows_differ(row, _row):
    """
    Check if grafana dashboard row and _row differ
    """
    row_copy = copy.deepcopy(row)
    _row_copy = copy.deepcopy(_row)
    # Strip id from all panels in both rows, since they are always generated.
    for panel in row_copy["panels"]:
        if "id" in panel:
            del panel["id"]
    for _panel in _row_copy["panels"]:
        if "id" in _panel:
            del _panel["id"]
    diff = DictDiffer(row_copy, _row_copy)
    return diff.changed() or diff.added() or diff.removed()


def dashboard_present(
    name,
    dashboard=None,
    dashboard_from_pillar=None,
    rows=None,
    rows_from_pillar=None,
    profile="grafana",
):
    """
    Ensure the grafana dashboard exists and is managed.

    name
        Name of the grafana dashboard.

    dashboard
        A dict that defines a dashboard that should be managed.

    dashboard_from_pillar
        A pillar key that contains a grafana dashboard dict. Mutually exclusive
        with dashboard.

    rows
        A list of grafana rows.

    rows_from_pillar
        A list of pillar keys that contain lists of grafana dashboard rows.
        Rows defined in the pillars will be appended to the rows defined in the
        state.

    profile
        A pillar key or dict that contains a list of hosts and an
        elasticsearch index to use.
    """
    ret = {"name": name, "result": None, "comment": "", "changes": {}}

    if not profile:
        raise SaltInvocationError("profile is a required argument.")
    if dashboard and dashboard_from_pillar:
        raise SaltInvocationError(
            "dashboard and dashboard_from_pillar are mutually exclusive arguments."
        )
    hosts, index = _parse_profile(profile)
    if not index:
        raise SaltInvocationError("index is a required key in the profile.")

    if not dashboard:
        dashboard = __salt__["pillar.get"](dashboard_from_pillar)
    if not rows:
        rows = []
    if rows_from_pillar:
        for key in rows_from_pillar:
            pillar_rows = __salt__["pillar.get"](key)
            # Pillar contains a list of rows
            if isinstance(pillar_rows, list):
                for row in pillar_rows:
                    rows.append(row)
            # Pillar contains a single row
            else:
                rows.append(pillar_rows)

    exists = __salt__["elasticsearch.exists"](
        index=index, id=name, doc_type="dashboard", hosts=hosts
    )

    if exists:
        _dashboard = __salt__["elasticsearch.get"](
            index=index, id=name, doc_type="dashboard", hosts=hosts
        )
        _dashboard = _dashboard.get("_source", {}).get("dashboard")
        _dashboard = salt.utils.json.loads(_dashboard)
    else:
        if not dashboard:
            raise SaltInvocationError(
                "Grafana dashboard does not exist and no"
                " dashboard template was provided."
            )
        if __opts__["test"]:
            ret["comment"] = f"Dashboard {name} is set to be created."
            ret["result"] = None
            return ret
        _dashboard = dashboard
    update_rows = []
    _ids = []
    _data = {}
    for _n, _row in enumerate(_dashboard["rows"]):
        # Collect the unique ids
        for _panel in _row["panels"]:
            if "id" in _panel:
                _ids.append(_panel["id"])
        # Collect all of the titles in the existing dashboard
        if "title" in _row:
            _data[_row["title"]] = _n
    _ids.sort()
    if not _ids:
        _ids = [1]
    for row in rows:
        if "title" not in row:
            raise SaltInvocationError("title is a required key for rows.")
        # Each panel needs to have a unique ID
        for panel in row["panels"]:
            _ids.append(_ids[-1] + 1)
            panel["id"] = _ids[-1]
        title = row["title"]
        # If the title doesn't exist, we need to add this row
        if title not in _data:
            update_rows.append(title)
            _dashboard["rows"].append(row)
            continue
        # For existing titles, replace the row if it differs
        _n = _data[title]
        if _rows_differ(row, _dashboard["rows"][_n]):
            _dashboard["rows"][_n] = row
            update_rows.append(title)
    if not update_rows:
        ret["result"] = True
        ret["comment"] = f"Dashboard {name} is up to date"
        return ret
    if __opts__["test"]:
        msg = f"Dashboard {name} is set to be updated."
        if update_rows:
            msg = f"{msg} The following rows set to be updated: {update_rows}"
        ret["comment"] = msg
        return ret
    body = {
        "user": "guest",
        "group": "guest",
        "title": name,
        "dashboard": salt.utils.json.dumps(_dashboard),
    }
    updated = __salt__["elasticsearch.index"](
        index=index, doc_type="dashboard", body=body, id=name, hosts=hosts
    )
    if updated:
        ret["result"] = True
        ret["changes"]["changed"] = name
        msg = f"Updated dashboard {name}."
        if update_rows:
            msg = f"{msg} The following rows were updated: {update_rows}"
        ret["comment"] = msg
    else:
        ret["result"] = False
        msg = f"Failed to update dashboard {name}."
        ret["comment"] = msg

    return ret


def dashboard_absent(name, hosts=None, profile="grafana"):
    """
    Ensure the named grafana dashboard is deleted.

    name
        Name of the grafana dashboard.

    profile
        A pillar key or dict that contains a list of hosts and an
        elasticsearch index to use.
    """
    ret = {"name": name, "result": None, "comment": "", "changes": {}}

    hosts, index = _parse_profile(profile)
    if not index:
        raise SaltInvocationError("index is a required key in the profile.")

    exists = __salt__["elasticsearch.exists"](
        index=index, id=name, doc_type="dashboard", hosts=hosts
    )

    if exists:
        if __opts__["test"]:
            ret["comment"] = f"Dashboard {name} is set to be removed."
            return ret
        deleted = __salt__["elasticsearch.delete"](
            index=index, doc_type="dashboard", id=name, hosts=hosts
        )
        if deleted:
            ret["result"] = True
            ret["changes"]["old"] = name
            ret["changes"]["new"] = None
        else:
            ret["result"] = False
            ret["comment"] = f"Failed to delete {name} dashboard."
    else:
        ret["result"] = True
        ret["comment"] = f"Dashboard {name} does not exist."

    return ret



import logging
import time

import attr
import pytest
from saltfactories.utils import random_string

log = logging.getLogger(__name__)


@attr.s(kw_only=True, slots=True)
class RabbitMQImage:
    name = attr.ib()
    tag = attr.ib()
    container_id = attr.ib()

    @container_id.default
    def _default_container_id(self):
        return random_string(f"{self.name}-{self.tag}-")

    def __str__(self):
        return f"{self.name}:{self.tag}"


@attr.s(kw_only=True, slots=True)
class RabbitMQCombo:
    rabbitmq_name = attr.ib()
    rabbitmq_version = attr.ib()


def get_test_versions():
    test_versions = []
    name = "rabbitmq"
    for version in (
        "3.9",
        "3.10",
    ):
        test_versions.append(
            RabbitMQImage(name=name, tag=version),
        )
    return test_versions


def get_test_version_id(value):
    return f"container={value}"


@pytest.fixture(scope="package", params=get_test_versions(), ids=get_test_version_id)
def rabbitmq_image(request):
    return request.param


def confirm_container_running(timeout_at, container):
    sleeptime = 1
    time.sleep(5)
    while time.time() <= timeout_at:
        ret = container.run("rabbitmqctl status --formatter=json")
        if ret.returncode == 0:
            break
        time.sleep(sleeptime)
        sleeptime *= 2
    else:
        return False
    return True


@pytest.fixture(scope="package")
def rabbitmq_container(salt_factories, rabbitmq_image):

    combo = RabbitMQCombo(
        rabbitmq_name=rabbitmq_image.name,
        rabbitmq_version=rabbitmq_image.tag,
    )
    container = salt_factories.get_container(
        rabbitmq_image.container_id,
        "ghcr.io/saltstack/salt-ci-containers/{}:{}".format(
            combo.rabbitmq_name, combo.rabbitmq_version
        ),
        container_run_kwargs={
            "ports": {"5672/tcp": None},
        },
        pull_before_start=True,
        skip_on_pull_failure=True,
        skip_if_docker_client_not_connectable=True,
    )
    container.container_start_check(confirm_container_running, container)
    with container.started():
        yield container


@pytest.fixture(scope="package")
def docker_cmd_run_all_wrapper(rabbitmq_container):
    def run_command(cmd, **kwargs):
        # Update rabbitmqctl location
        if cmd[0] is None:
            cmd[0] = "/opt/rabbitmq/sbin/rabbitmqctl"

        ret = rabbitmq_container.run(cmd)
        res = {"retcode": ret.returncode, "stdout": ret.stdout, "stderr": ret.stderr}
        return res

    return run_command


@pytest.fixture(scope="package")
def docker_cmd_run_wrapper(rabbitmq_container):
    def run_command(cmd, **kwargs):
        # Update rabbitmqctl location
        if cmd[0] is None:
            cmd[0] = "/opt/rabbitmq/sbin/rabbitmqctl"

        ret = rabbitmq_container.run(cmd)
        return ret.stdout

    return run_command



"""
Package support for AIX

.. important::
    If you feel that Salt should be using this module to manage filesets or
    rpm packages on a minion, and it is using a different module (or gives an
    error similar to *'pkg.install' is not available*), see :ref:`here
    <module-provider-override>`.
"""

import copy
import logging
import os
import pathlib

import salt.utils.data
import salt.utils.functools
import salt.utils.path
import salt.utils.pkg
from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "pkg"


def __virtual__():
    """
    Set the virtual pkg module if the os is AIX
    """
    if __grains__["os_family"] == "AIX":
        return __virtualname__
    return (False, "Did not load AIX module on non-AIX OS.")


def _check_pkg(target):
    """
    Return name, version and if rpm package for specified target
    """
    ret = {}
    cmd = ["/usr/bin/lslpp", "-Lc", target]
    result = __salt__["cmd.run_all"](cmd, python_shell=False)

    if 0 == result["retcode"]:
        name = ""
        version_num = ""
        rpmpkg = False
        lines = result["stdout"].splitlines()
        for line in lines:
            if line.startswith("#"):
                continue

            comps = line.split(":")
            if len(comps) < 7:
                raise CommandExecutionError(
                    "Error occurred finding fileset/package",
                    info={"errors": comps[1].strip()},
                )

            # handle first matching line
            if "R" in comps[6]:
                name = comps[0]
                rpmpkg = True
            else:
                name = comps[1]  # use fileset rather than rpm package

            version_num = comps[2]
            break

        return name, version_num, rpmpkg
    else:
        raise CommandExecutionError(
            "Error occurred finding fileset/package",
            info={"errors": result["stderr"].strip()},
        )


def _is_installed_rpm(name):
    """
    Returns True if the rpm package is installed. Otherwise returns False.
    """
    cmd = ["/usr/bin/rpm", "-q", name]
    return __salt__["cmd.retcode"](cmd) == 0


def _list_pkgs_from_context(versions_as_list):
    """
    Use pkg list from __context__
    """
    if versions_as_list:
        return __context__["pkg.list_pkgs"]
    else:
        ret = copy.deepcopy(__context__["pkg.list_pkgs"])
        __salt__["pkg_resource.stringify"](ret)
        return ret


def list_pkgs(versions_as_list=False, **kwargs):
    """
    List the filesets/rpm packages currently installed as a dict:

    .. code-block:: python

        {'<package_name>': '<version>'}

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.list_pkgs
    """
    ret = {}
    versions_as_list = salt.utils.data.is_true(versions_as_list)
    # not yet implemented or not applicable
    if any(
        [salt.utils.data.is_true(kwargs.get(x)) for x in ("removed", "purge_desired")]
    ):
        return ret

    if "pkg.list_pkgs" in __context__ and kwargs.get("use_context", True):
        return _list_pkgs_from_context(versions_as_list)

    # cmd returns information colon delimited in a single linei, format
    #   Package Name:Fileset:Level:State:PTF Id:Fix State:Type:Description:
    #       Destination Dir.:Uninstaller:Message Catalog:Message Set:
    #       Message Number:Parent:Automatic:EFIX Locked:Install Path:Build Date
    # Example:
    #   xcursor:xcursor-1.1.7-3:1.1.7-3: : :C:R:X Cursor library: :\
    #       /bin/rpm -e xcursor: : : : :0: :(none):Mon May  8 15:18:35 CDT 2017
    #   bos:bos.rte.libcur:7.1.5.0: : :C:F:libcurses Library: : : : : : :0:0:/:1731
    #
    # where Type codes: F -- Installp Fileset, P -- Product, C -- Component,
    #                   T -- Feature, R -- RPM Package
    cmd = "/usr/bin/lslpp -Lc"
    lines = __salt__["cmd.run"](cmd, python_shell=False).splitlines()

    for line in lines:
        if line.startswith("#"):
            continue

        comps = line.split(":")
        if len(comps) < 7:
            continue

        if "R" in comps[6]:
            name = comps[0]
        else:
            name = comps[1]  # use fileset rather than rpm package

        version_num = comps[2]
        __salt__["pkg_resource.add_pkg"](ret, name, version_num)

    __salt__["pkg_resource.sort_pkglist"](ret)
    __context__["pkg.list_pkgs"] = copy.deepcopy(ret)

    if not versions_as_list:
        __salt__["pkg_resource.stringify"](ret)

    return ret


def version(*names, **kwargs):
    """
    Return the current installed version of the named fileset/rpm package
    If more than one fileset/rpm package name is specified a dict of
    name/version pairs is returned.

    .. versionchanged:: 3005

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.latest_version <package name>
        salt '*' pkg.latest_version <package1> <package2> <package3> ...

    """
    kwargs.pop("refresh", True)

    ret = {}
    if not names:
        return ""
    for name in names:
        # AIX packaging includes info on filesets and rpms
        version_found = ""
        cmd = f"lslpp -Lq {name}"
        aix_info = __salt__["cmd.run_all"](cmd, python_shell=False)
        if 0 == aix_info["retcode"]:
            aix_info_list = aix_info["stdout"].split("\n")
            log.debug(
                "Returned AIX packaging information aix_info_list %s for name %s",
                aix_info_list,
                name,
            )
            for aix_line in aix_info_list:
                if name in aix_line:
                    aix_ver_list = aix_line.split()
                    log.debug(
                        "Processing name %s with AIX packaging version information %s",
                        name,
                        aix_ver_list,
                    )
                    version_found = aix_ver_list[1]
                    if version_found:
                        log.debug(
                            "Found name %s in AIX packaging information, version %s",
                            name,
                            version_found,
                        )
                        break
        else:
            log.debug("Could not find name %s in AIX packaging information", name)

        ret[name] = version_found

    # Return a string if only one package name passed
    if len(names) == 1:
        return ret[names[0]]
    return ret


def _is_installed(name, **kwargs):
    """
    Returns True if the fileset/rpm package is installed. Otherwise returns False.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg._is_installed bash
    """
    cmd = ["/usr/bin/lslpp", "-Lc", name]
    return __salt__["cmd.retcode"](cmd) == 0


def install(name=None, refresh=False, pkgs=None, version=None, test=False, **kwargs):
    """
    Install the named fileset(s)/rpm package(s).

    .. versionchanged:: 3005

        preference to install rpm packages are to use in the following order:
            /opt/freeware/bin/dnf
            /opt/freeware/bin/yum
            /usr/bin/yum
            /usr/bin/rpm

    .. note:
        use of rpm to install implies that rpm's dependencies must have been previously installed.
        dnf and yum automatically install rpm's dependencies as part of the install process

        Alogrithm to install filesets or rpms is as follows:
            if ends with '.rte' or '.bff'
                process as fileset
            if ends with '.rpm'
                process as rpm
            if unrecognised or no file extension
                attempt process with dnf | yum
                failure implies attempt process as fileset

        Fileset needs to be available as a single path and filename
        compound filesets are not handled and are not supported.
        An example is bos.adt.insttools which is part of bos.adt.other and is installed as follows
        /usr/bin/installp -acXYg /cecc/repos/aix72/TL4/BASE/installp/ppc/bos.adt.other bos.adt.insttools

    name
        The name of the fileset or rpm package to be installed.

    refresh
        Whether or not to update the yum database before executing.


    pkgs
        A list of filesets and/or rpm packages to install.
        Must be passed as a python list. The ``name`` parameter will be
        ignored if this option is passed.

    version
        Install a specific version of a fileset/rpm package.
        (Unused at present).

        .. note::
            Remember that versions that contain a single `.` will be interpreted
            as numbers and must be double-quoted. For example, version
            ``3006.10`` will be rendered as ``3006.1``. To pass ``3006.10``
            you'll need to use double-quotes. ``version="'3006.10'"``

    test
        Verify that command functions correctly.

    Returns a dict containing the new fileset(s)/rpm package(s) names and versions:

        {'<package>': {'old': '<old-version>',
                       'new': '<new-version>'}}

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.install /stage/middleware/AIX/bash-4.2-3.aix6.1.ppc.rpm
        salt '*' pkg.install /stage/middleware/AIX/bash-4.2-3.aix6.1.ppc.rpm refresh=True
        salt '*' pkg.install /stage/middleware/AIX/VIOS2211_update/tpc_4.1.1.85.bff
        salt '*' pkg.install /cecc/repos/aix72/TL3/BASE/installp/ppc/bos.rte.printers_7.2.2.0.bff
        salt '*' pkg.install /stage/middleware/AIX/Xlc/usr/sys/inst.images/xlC.rte
        salt '*' pkg.install /stage/middleware/AIX/Firefox/ppc-AIX53/Firefox.base
        salt '*' pkg.install /cecc/repos/aix72/TL3/BASE/installp/ppc/bos.net
        salt '*' pkg.install pkgs='["foo", "bar"]'
        salt '*' pkg.install libxml2
    """
    targets = salt.utils.args.split_input(pkgs) if pkgs else [name]
    if not targets:
        return {}

    if pkgs:
        log.debug("Installing these fileset(s)/rpm package(s) %s: %s", name, targets)

    # Get a list of the currently installed pkgs.
    old = list_pkgs()

    # Install the fileset (normally ends with bff or rte) or rpm package(s)
    errors = []
    for target in targets:
        filename = os.path.basename(target)
        flag_fileset = False
        flag_actual_rpm = False
        flag_try_rpm_failed = False
        cmd = ""
        out = {}
        if filename.endswith(".bff") or filename.endswith(".rte"):
            flag_fileset = True
            log.debug("install identified %s as fileset", filename)
        else:
            if filename.endswith(".rpm"):
                flag_actual_rpm = True
                log.debug("install identified %s as rpm", filename)
            else:
                log.debug("install filename %s trying install as rpm", filename)

            # assume use dnf or yum
            cmdflags = "install "
            libpathenv = {"LIBPATH": "/opt/freeware/lib:/usr/lib"}
            if pathlib.Path("/opt/freeware/bin/dnf").is_file():
                cmdflags += "--allowerasing "
                cmdexe = "/opt/freeware/bin/dnf"
                if test:
                    cmdflags += "--assumeno "
                else:
                    cmdflags += "--assumeyes "
                if refresh:
                    cmdflags += "--refresh "

                cmd = f"{cmdexe} {cmdflags} {target}"
                out = __salt__["cmd.run_all"](
                    cmd,
                    python_shell=False,
                    env=libpathenv,
                    ignore_retcode=True,
                )

            elif pathlib.Path("/usr/bin/yum").is_file():
                # check for old yum first, removed if new dnf or yum
                cmdexe = "/usr/bin/yum"
                if test:
                    cmdflags += "--assumeno "
                else:
                    cmdflags += "--assumeyes "

                cmd = f"{cmdexe} {cmdflags} {target}"
                out = __salt__["cmd.run_all"](
                    cmd,
                    python_shell=False,
                    env=libpathenv,
                    ignore_retcode=True,
                )

            elif pathlib.Path("/opt/freeware/bin/yum").is_file():
                cmdflags += "--allowerasing "
                cmdexe = "/opt/freeware/bin/yum"
                if test:
                    cmdflags += "--assumeno "
                else:
                    cmdflags += "--assumeyes "
                if refresh:
                    cmdflags += "--refresh "

                cmd = f"{cmdexe} {cmdflags} {target}"
                out = __salt__["cmd.run_all"](
                    cmd,
                    python_shell=False,
                    env=libpathenv,
                    ignore_retcode=True,
                )

            else:
                cmdexe = "/usr/bin/rpm"
                cmdflags = "-Uivh "
                if test:
                    cmdflags += "--test"

                cmd = f"{cmdexe} {cmdflags} {target}"
                out = __salt__["cmd.run_all"](cmd, python_shell=False)

        if "retcode" in out and not (0 == out["retcode"] or 100 == out["retcode"]):
            if not flag_actual_rpm:
                flag_try_rpm_failed = True
                log.debug(
                    "install tried filename %s as rpm and failed, trying as fileset",
                    filename,
                )
            else:
                errors.append(out["stderr"])
                log.debug(
                    "install error rpm path, returned result %s, resultant errors %s",
                    out,
                    errors,
                )

        if flag_fileset or flag_try_rpm_failed:
            # either identified as fileset, or failed trying install as rpm, try as fileset

            cmd = "/usr/sbin/installp -acYXg"
            if test:
                cmd += "p"
            cmd += " -d "
            dirpath = os.path.dirname(target)
            cmd += dirpath + " " + filename
            log.debug("install fileset commanda to attempt %s", cmd)
            out = __salt__["cmd.run_all"](cmd, python_shell=False)
            if 0 != out["retcode"]:
                errors.append(out["stderr"])
                log.debug(
                    "install error fileset path, returned result %s, resultant errors %s",
                    out,
                    errors,
                )

    # Get a list of the packages after the uninstall
    __context__.pop("pkg.list_pkgs", None)
    new = list_pkgs()
    ret = salt.utils.data.compare_dicts(old, new)

    if errors:
        raise CommandExecutionError(
            "Problems encountered installing filesets(s)/package(s)",
            info={"changes": ret, "errors": errors},
        )

    # No error occurred
    if test:
        return "Test succeeded."

    return ret


def remove(name=None, pkgs=None, **kwargs):
    """
    Remove specified fileset(s)/rpm package(s).

    name
        The name of the fileset or rpm package to be deleted.

    .. versionchanged:: 3005

        preference to install rpm packages are to use in the following order:
            /opt/freeware/bin/dnf
            /opt/freeware/bin/yum
            /usr/bin/yum
            /usr/bin/rpm

    pkgs
        A list of filesets and/or rpm packages to delete.
        Must be passed as a python list. The ``name`` parameter will be
        ignored if this option is passed.


    Returns a list containing the removed packages.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.remove <fileset/rpm package name>
        salt '*' pkg.remove tcsh
        salt '*' pkg.remove xlC.rte
        salt '*' pkg.remove Firefox.base.adt
        salt '*' pkg.remove pkgs='["foo", "bar"]'
    """
    targets = salt.utils.args.split_input(pkgs) if pkgs else [name]
    if not targets:
        return {}

    if pkgs:
        log.debug("Removing these fileset(s)/rpm package(s) %s: %s", name, targets)

    errors = []

    # Get a list of the currently installed pkgs.
    old = list_pkgs()

    # Remove the fileset or rpm package(s)
    for target in targets:
        cmd = ""
        out = {}
        try:
            named, versionpkg, rpmpkg = _check_pkg(target)
        except CommandExecutionError as exc:
            if exc.info:
                errors.append(exc.info["errors"])
            continue

        if rpmpkg:

            # assume use dnf or yum
            cmdflags = "-y remove"
            libpathenv = {"LIBPATH": "/opt/freeware/lib:/usr/lib"}
            if pathlib.Path("/opt/freeware/bin/dnf").is_file():
                cmdexe = "/opt/freeware/bin/dnf"
                cmd = f"{cmdexe} {cmdflags} {target}"
                out = __salt__["cmd.run_all"](
                    cmd,
                    python_shell=False,
                    env=libpathenv,
                    ignore_retcode=True,
                )
            elif pathlib.Path("/opt/freeware/bin/yum").is_file():
                cmdexe = "/opt/freeware/bin/yum"
                cmd = f"{cmdexe} {cmdflags} {target}"
                out = __salt__["cmd.run_all"](
                    cmd,
                    python_shell=False,
                    env=libpathenv,
                    ignore_retcode=True,
                )
            elif pathlib.Path("/usr/bin/yum").is_file():
                cmdexe = "/usr/bin/yum"
                cmd = f"{cmdexe} {cmdflags} {target}"
                out = __salt__["cmd.run_all"](
                    cmd,
                    python_shell=False,
                    env=libpathenv,
                    ignore_retcode=True,
                )
            else:
                cmdexe = "/usr/bin/rpm"
                cmdflags = "-e"
                cmd = f"{cmdexe} {cmdflags} {target}"
                out = __salt__["cmd.run_all"](cmd, python_shell=False)
        else:
            cmd = ["/usr/sbin/installp", "-u", named]
            out = __salt__["cmd.run_all"](cmd, python_shell=False)

        log.debug("result of removal command %s, returned result %s", cmd, out)

    # Get a list of the packages after the uninstall
    __context__.pop("pkg.list_pkgs", None)
    new = list_pkgs()
    ret = salt.utils.data.compare_dicts(old, new)

    if errors:
        raise CommandExecutionError(
            "Problems encountered removing filesets(s)/package(s)",
            info={"changes": ret, "errors": errors},
        )

    return ret


def latest_version(*names, **kwargs):
    """
    Return the latest available version of the named fileset/rpm package available for
    upgrade or installation. If more than one fileset/rpm package name is
    specified, a dict of name/version pairs is returned.

    If the latest version of a given fileset/rpm package is already installed,
    an empty string will be returned for that package.

    .. versionchanged:: 3005

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.latest_version <package name>
        salt '*' pkg.latest_version <package1> <package2> <package3> ...

    Note: currently only functional for rpm packages due to filesets do not have a specific location to check
        Requires yum of dnf available in order to query a repository

    This function will always return an empty string for unfound fileset/rpm package.
    """
    kwargs.pop("refresh", True)

    ret = {}
    if not names:
        return ""
    for name in names:
        # AIX packaging includes info on filesets and rpms
        version_found = ""
        libpathenv = {"LIBPATH": "/opt/freeware/lib:/usr/lib"}
        if pathlib.Path("/opt/freeware/bin/dnf").is_file():
            cmdexe = "/opt/freeware/bin/dnf"
            cmd = f"{cmdexe} check-update {name}"
            available_info = __salt__["cmd.run_all"](
                cmd, python_shell=False, env=libpathenv, ignore_retcode=True
            )
        elif pathlib.Path("/opt/freeware/bin/yum").is_file():
            cmdexe = "/opt/freeware/bin/yum"
            cmd = f"{cmdexe} check-update {name}"
            available_info = __salt__["cmd.run_all"](
                cmd, python_shell=False, env=libpathenv, ignore_retcode=True
            )
        elif pathlib.Path("/usr/bin/yum").is_file():
            cmdexe = "/usr/bin/yum"
            cmd = f"{cmdexe} check-update {name}"
            available_info = __salt__["cmd.run_all"](
                cmd, python_shell=False, env=libpathenv, ignore_retcode=True
            )
        else:
            # no yum found implies no repository support
            available_info = None

        log.debug(
            "latest_version dnf|yum check-update command returned information %s",
            available_info,
        )
        if available_info and (
            0 == available_info["retcode"] or 100 == available_info["retcode"]
        ):
            available_output = available_info["stdout"]
            if available_output:
                available_list = available_output.split()
                flag_found = False
                for name_chk in available_list:
                    # have viable check, note .ppc or .noarch
                    if name_chk.startswith(name):
                        # check full name
                        pkg_label = name_chk.split(".")
                        if name == pkg_label[0]:
                            flag_found = True
                    elif flag_found:
                        # version comes after name found
                        version_found = name_chk
                        break

        if version_found:
            log.debug(
                "latest_version result for name %s found version %s",
                name,
                version_found,
            )
        else:
            log.debug("Could not find AIX / RPM packaging version for %s", name)

        ret[name] = version_found

    # Return a string if only one package name passed
    if len(names) == 1:
        return ret[names[0]]
    return ret


# available_version is being deprecated
available_version = salt.utils.functools.alias_function(
    latest_version, "available_version"
)


def upgrade_available(name, **kwargs):
    """
    Check whether or not an upgrade is available for a given package

    .. versionchanged:: 3005

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.upgrade_available <package name>

    Note: currently only functional for rpm packages due to filesets do not have a specific location to check
        Requires yum of dnf available in order to query a repository

    """
    # AIX packaging includes info on filesets and rpms
    rpm_found = False
    version_found = ""

    libpathenv = {"LIBPATH": "/opt/freeware/lib:/usr/lib"}
    if pathlib.Path("/opt/freeware/bin/dnf").is_file():
        cmdexe = "/opt/freeware/bin/dnf"
        cmd = f"{cmdexe} check-update {name}"
        available_info = __salt__["cmd.run_all"](
            cmd, python_shell=False, env=libpathenv, ignore_retcode=True
        )
    elif pathlib.Path("/opt/freeware/bin/yum").is_file():
        cmdexe = "/opt/freeware/bin/yum"
        cmd = f"{cmdexe} check-update {name}"
        available_info = __salt__["cmd.run_all"](
            cmd, python_shell=False, env=libpathenv, ignore_retcode=True
        )
    elif pathlib.Path("/usr/bin/yum").is_file():
        cmdexe = "/usr/bin/yum"
        cmd = f"{cmdexe} check-update {name}"
        available_info = __salt__["cmd.run_all"](
            cmd, python_shell=False, env=libpathenv, ignore_retcode=True
        )
    else:
        # no yum found implies no repository support
        return False

    log.debug(
        "upgrade_available yum check-update command %s, returned information %s",
        cmd,
        available_info,
    )
    if 0 == available_info["retcode"] or 100 == available_info["retcode"]:
        available_output = available_info["stdout"]
        if available_output:
            available_list = available_output.split()
            flag_found = False
            for name_chk in available_list:
                # have viable check, note .ppc or .noarch
                if name_chk.startswith(name):
                    # check full name
                    pkg_label = name_chk.split(".")
                    if name == pkg_label[0]:
                        flag_found = True
                elif flag_found:
                    # version comes after name found
                    version_found = name_chk
                    break

        current_version = version(name)
        log.debug(
            "upgrade_available result for name %s, found current version %s, available version %s",
            name,
            current_version,
            version_found,
        )

    if version_found:
        return current_version != version_found
    else:
        log.debug("upgrade_available information for name %s was not found", name)
        return False



"""
integration tests for mac_power
"""

import pytest

from salt.exceptions import CommandExecutionError, SaltInvocationError

pytestmark = [
    pytest.mark.skip_if_binaries_missing("systemsetup"),
    pytest.mark.slow_test,
    pytest.mark.destructive_test,
    pytest.mark.skip_if_not_root,
    pytest.mark.skip_unless_on_darwin,
]


@pytest.fixture(scope="module")
def power(modules):
    return modules.power


@pytest.fixture
def _reset_computer_sleep(power):
    ret = power.get_computer_sleep()
    try:
        yield
    finally:
        power.set_computer_sleep(ret)


@pytest.fixture
def _reset_display_sleep(power):
    ret = power.get_display_sleep()
    try:
        yield
    finally:
        power.set_display_sleep(ret)


@pytest.fixture
def _reset_harddisk_sleep(power):
    ret = power.get_harddisk_sleep()
    try:
        yield
    finally:
        power.set_harddisk_sleep(ret)


@pytest.fixture
def _reset_restart_power_failure(power):
    try:
        ret = power.get_restart_power_failure()
        if not isinstance(ret, bool):
            assert "Error" in ret
            pytest.skip(f"Error while calling `get_restart_power_failure()`: {ret}")
    except CommandExecutionError as exc:
        if "Not supported on this machine" in str(exc):
            pytest.skip("Restart After Power Failure: Not supported on this machine.")
    try:
        yield
    finally:
        if isinstance(ret, bool):
            if ret:
                ret = power.set_restart_power_failure("On")
                assert ret
            else:
                ret = power.set_restart_power_failure("Off")
                assert ret


@pytest.fixture
def _reset_sleep_on_power_button(power):
    try:
        ret = power.get_sleep_on_power_button()
        if not isinstance(ret, bool):
            functionality_available = False
        else:
            functionality_available = True
    except CommandExecutionError as exc:
        functionality_available = False

    if functionality_available is False:
        pytest.skip("Skipping. sleep_on_power_button unavailable.")

    try:
        yield
    finally:
        power.set_sleep_on_power_button(ret)


@pytest.fixture
def _reset_wake_on_modem(power):
    try:
        ret = power.get_wake_on_modem()
        if not isinstance(ret, bool):
            functionality_available = False
        else:
            functionality_available = True
    except CommandExecutionError as exc:
        functionality_available = False

    if functionality_available is False:
        pytest.skip("Skipping. wake_on_modem unavailable.")

    try:
        yield
    finally:
        power.set_wake_on_modem(ret)


@pytest.fixture
def _reset_wake_on_network(power):
    try:
        ret = power.get_wake_on_network()
        if not isinstance(ret, bool):
            assert "Error" in ret
            pytest.skip(f"Error while calling `get_wake_on_network()`: {ret}")
    except CommandExecutionError as exc:
        if "Not supported on this machine" in str(exc):
            pytest.skip("Wake On Network Access: Not supported on this machine")
    try:
        yield
    finally:
        if isinstance(ret, bool):
            ret = power.set_wake_on_network(ret)
            assert ret


@pytest.mark.usefixtures("_reset_computer_sleep")
def test_computer_sleep(power):
    """
    Test power.get_computer_sleep
    Test power.set_computer_sleep
    """

    # Normal Functionality
    ret = power.set_computer_sleep(90)
    assert ret

    ret = power.get_computer_sleep()
    assert ret == "after 90 minutes"

    ret = power.set_computer_sleep("Off")
    assert ret

    ret = power.get_computer_sleep()
    assert ret == "Never"

    # Test invalid input
    with pytest.raises(SaltInvocationError) as exc:
        power.set_computer_sleep("spongebob")
        assert "Invalid String Value for Minutes" in str(exc.value)

    with pytest.raises(SaltInvocationError) as exc:
        power.set_computer_sleep(0)
        assert "Invalid Integer Value for Minutes" in str(exc.value)

    with pytest.raises(SaltInvocationError) as exc:
        power.set_computer_sleep(181)
        assert "Invalid Integer Value for Minutes" in str(exc.value)

    with pytest.raises(SaltInvocationError) as exc:
        power.set_computer_sleep(True)
        assert "Invalid Boolean Value for Minutes" in str(exc.value)


@pytest.mark.usefixtures("_reset_display_sleep")
def test_display_sleep(power):
    """
    Test power.get_display_sleep
    Test power.set_display_sleep
    """

    # Normal Functionality
    ret = power.set_display_sleep(90)
    assert ret

    ret = power.get_display_sleep()
    assert ret == "after 90 minutes"

    ret = power.set_display_sleep("Off")
    assert ret

    ret = power.get_display_sleep()
    assert ret == "Never"

    # Test invalid input
    with pytest.raises(SaltInvocationError) as exc:
        power.set_display_sleep("spongebob")
        assert "Invalid String Value for Minutes" in str(exc.value)

    with pytest.raises(SaltInvocationError) as exc:
        power.set_display_sleep(0)
        assert "Invalid Integer Value for Minutes" in str(exc.value)

    with pytest.raises(SaltInvocationError) as exc:
        power.set_display_sleep(181)
        assert "Invalid Integer Value for Minutes" in str(exc.value)

    with pytest.raises(SaltInvocationError) as exc:
        power.set_display_sleep(True)
        assert "Invalid Boolean Value for Minutes" in str(exc.value)


@pytest.mark.usefixtures("_reset_harddisk_sleep")
def test_harddisk_sleep(power):
    """
    Test power.get_harddisk_sleep
    Test power.set_harddisk_sleep
    """

    # Normal Functionality
    ret = power.set_harddisk_sleep(90)
    assert ret

    ret = power.get_harddisk_sleep()
    assert ret == "after 90 minutes"

    ret = power.set_harddisk_sleep("Off")
    assert ret

    ret = power.get_harddisk_sleep()
    assert ret == "Never"

    # Test invalid input
    with pytest.raises(SaltInvocationError) as exc:
        power.set_harddisk_sleep("spongebob")
        assert "Invalid String Value for Minutes" in str(exc.value)

    with pytest.raises(SaltInvocationError) as exc:
        power.set_harddisk_sleep(0)
        assert "Invalid Integer Value for Minutes" in str(exc.value)

    with pytest.raises(SaltInvocationError) as exc:
        power.set_harddisk_sleep(181)
        assert "Invalid Integer Value for Minutes" in str(exc.value)

    with pytest.raises(SaltInvocationError) as exc:
        power.set_harddisk_sleep(True)
        assert "Invalid Boolean Value for Minutes" in str(exc.value)


def test_restart_freeze(power):
    """
    Test power.get_restart_freeze
    Test power.set_restart_freeze
    """
    # Normal Functionality
    ret = power.set_restart_freeze("on")
    assert ret

    ret = power.get_restart_freeze()
    assert ret

    # This will return False because mac fails to actually make the change
    ret = power.set_restart_freeze("off")
    assert not ret

    # Even setting to off returns true, it actually is never set
    # This is an apple bug
    ret = power.get_restart_freeze()
    assert ret


@pytest.mark.usefixtures("_reset_restart_power_failure")
def test_restart_power_failure(power):
    """
    Test power.get_restart_power_failure
    Test power.set_restart_power_failure
    """
    ret = power.set_restart_power_failure("On")
    assert ret

    ret = power.get_restart_power_failure()
    assert ret

    ret = power.set_restart_power_failure("Off")
    assert ret

    ret = power.get_restart_power_failure()
    assert not ret


@pytest.mark.usefixtures("_reset_sleep_on_power_button")
def test_sleep_on_power_button(power):
    """
    Test power.get_sleep_on_power_button
    Test power.set_sleep_on_power_button
    """
    ret = power.set_sleep_on_power_button("on")
    assert ret

    ret = power.get_sleep_on_power_button()
    assert ret

    ret = power.set_sleep_on_power_button("off")
    assert ret

    ret = power.get_sleep_on_power_button()
    assert not ret


@pytest.mark.usefixtures("_reset_wake_on_modem")
def test_wake_on_modem(power):
    """
    Test power.get_wake_on_modem
    Test power.set_wake_on_modem
    """
    ret = power.set_wake_on_modem("on")
    assert ret

    ret = power.get_wake_on_modem()
    assert ret

    ret = power.set_wake_on_modem("off")
    assert ret

    ret = power.get_wake_on_modem()
    assert not ret


@pytest.mark.usefixtures("_reset_wake_on_network")
def test_wake_on_network(power):
    """
    Test power.get_wake_on_network
    Test power.set_wake_on_network
    """
    ret = power.set_wake_on_network("on")
    assert ret

    ret = power.get_wake_on_network()
    assert ret

    ret = power.set_wake_on_network("off")
    assert ret

    ret = power.get_wake_on_network()
    assert not ret



import logging

import pytest
from saltfactories.utils.tempfiles import temp_directory

import salt.client.ssh.__init__ as dunder_ssh
from salt.exceptions import SaltClientError, SaltSystemExit
from tests.support.mock import MagicMock, patch

pytestmark = [pytest.mark.skip_unless_on_linux(reason="Test ssh only run on Linux")]


log = logging.getLogger(__name__)


def test_salt_refs():
    data_strg_cats = "cats"
    ret = dunder_ssh.salt_refs(data_strg_cats)
    assert ret == []

    data_strg_proto = "salt://test_salt_ref"
    ret = dunder_ssh.salt_refs(data_strg_proto)
    assert ret == [data_strg_proto]

    data_list_no_proto = ["cats"]
    ret = dunder_ssh.salt_refs(data_list_no_proto)
    assert ret == []

    data_list_proto = ["salt://test_salt_ref1", "salt://test_salt_ref2", "cats"]
    ret = dunder_ssh.salt_refs(data_list_proto)
    assert ret == ["salt://test_salt_ref1", "salt://test_salt_ref2"]


def test_convert_args():
    test_args = [
        "arg1",
        {"key1": "value1", "key2": "value2", "__kwarg__": "kwords"},
        "dog1",
    ]
    expected = ["arg1", "key1=value1", "key2=value2", "dog1"]
    ret = dunder_ssh._convert_args(test_args)
    assert ret == expected


def test_ssh_class():

    with temp_directory() as temp_dir:
        assert temp_dir.is_dir()
        opts = {
            "sock_dir": temp_dir,
            "regen_thin": False,
            "__master_opts__": {"pki_dir": "pki"},
            "selected_target_option": None,
            "tgt": "*",
            "tgt_type": "glob",
            "fileserver_backend": ["roots"],
            "cachedir": "/tmp",
            "thin_extra_mods": "",
            "ssh_ext_alternatives": None,
        }

        with patch("salt.utils.path.which", return_value=""), pytest.raises(
            SaltSystemExit
        ) as err:
            test_ssh = dunder_ssh.SSH(opts)
            assert (
                "salt-ssh could not be run because it could not generate keys."
                in str(err.value)
            )

        with patch("salt.utils.path.which", return_value="/usr/bin/ssh"), patch(
            "os.path.isfile", return_value=False
        ), patch(
            "salt.client.ssh.shell.gen_key", MagicMock(side_effect=OSError())
        ), pytest.raises(
            SaltClientError
        ) as err:
            test_ssh = dunder_ssh.SSH(opts)
            assert (
                "salt-ssh could not be run because it could not generate keys."
                in err.value
            )



import logging

import pytest

from salt.modules.virtualenv_mod import KNOWN_BINARY_NAMES
from tests.support.helpers import SKIP_INITIAL_PHOTONOS_FAILURES

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_on_fips_enabled_platform,
    pytest.mark.skip_if_binaries_missing(*KNOWN_BINARY_NAMES, check_all=False),
]


def _check_skip(grains):
    if grains["os"] == "MacOS":
        return True
    return False


@SKIP_INITIAL_PHOTONOS_FAILURES
@pytest.mark.destructive_test
@pytest.mark.skip_if_not_root
@pytest.mark.skip_initial_gh_actions_failure(skip=_check_skip)
def test_issue_1959_virtualenv_runas(tmp_path_world_rw, state_tree, states):
    with pytest.helpers.create_account(create_group=True) as account:

        state_tree_dirname = account.username
        state_tree_path = state_tree / state_tree_dirname
        with pytest.helpers.temp_file(
            "requirements.txt", contents="pep8==1.3.3\n", directory=state_tree_path
        ):
            venv_dir = tmp_path_world_rw / "venv"
            ret = states.virtualenv.managed(
                name=str(venv_dir),
                user=account.username,
                requirements=f"salt://{state_tree_dirname}/requirements.txt",
            )
            assert ret.result is True

            # Lets check proper ownership
            statinfo = venv_dir.stat()
            assert statinfo.st_uid == account.info.uid


@pytest.mark.parametrize("requirement", ["pep8==1.3.3", "zope.interface==5.0.0"])
def test_issue_2594_non_invalidated_cache(tmp_path, state_tree, modules, requirement):
    state_tree_dirname = "issue-2594"
    state_tree_path = state_tree / state_tree_dirname
    with pytest.helpers.temp_file(
        "requirements.txt", contents=requirement, directory=state_tree_path
    ):
        venv_dir = tmp_path / "venv"

        # Our state template
        template = [
            f"{venv_dir}:",
            "  virtualenv.managed:",
            "    - system_site_packages: False",
            "    - clear: false",
            f"    - requirements: salt://{state_tree_dirname}/requirements.txt",
        ]

        # Let's run our state!!!
        ret = modules.state.template_str("\n".join(template))
        assert not ret.failed
        for entry in ret:
            assert entry.result is True
            assert "Created new virtualenv" in entry.comment
            assert "packages" in entry.changes
            assert "new" in entry.changes["packages"]
            assert requirement in entry.changes["packages"]["new"]

        # Let's make sure, it really got installed
        ret = modules.pip.freeze(bin_env=str(venv_dir))
        assert requirement in ret



"""
    Unit tests for salt.utils.yamldumper
"""

from collections import OrderedDict, defaultdict

import salt.utils.yamldumper
from salt.utils.context import NamespacedDictWrapper
from salt.utils.datastructures import HashableOrderedDict


def test_yaml_dump():
    """
    Test yaml.dump a dict
    """
    data = {"foo": "bar"}
    exp_yaml = "{foo: bar}\n"

    assert salt.utils.yamldumper.dump(data) == exp_yaml

    assert salt.utils.yamldumper.dump(
        data, default_flow_style=False
    ) == exp_yaml.replace("{", "").replace("}", "")


def test_yaml_safe_dump():
    """
    Test yaml.safe_dump a dict
    """
    data = {"foo": "bar"}
    assert salt.utils.yamldumper.safe_dump(data) == "{foo: bar}\n"

    assert (
        salt.utils.yamldumper.safe_dump(data, default_flow_style=False) == "foo: bar\n"
    )


def test_yaml_ordered_dump():
    """
    Test yaml.dump with OrderedDict
    """
    data = OrderedDict([("foo", "bar"), ("baz", "qux")])
    exp_yaml = "{foo: bar, baz: qux}\n"
    assert (
        salt.utils.yamldumper.dump(data, Dumper=salt.utils.yamldumper.OrderedDumper)
        == exp_yaml
    )


def test_yaml_safe_ordered_dump():
    """
    Test yaml.safe_dump with OrderedDict
    """
    data = OrderedDict([("foo", "bar"), ("baz", "qux")])
    exp_yaml = "{foo: bar, baz: qux}\n"
    assert salt.utils.yamldumper.safe_dump(data) == exp_yaml


def test_yaml_indent_safe_ordered_dump():
    """
    Test yaml.dump with IndentedSafeOrderedDumper
    """
    data = OrderedDict([("foo", ["bar", "baz"]), ("qux", "quux")])
    # Account for difference in SafeDumper vs CSafeDumper
    if salt.utils.yamldumper.SafeDumper.__name__ == "SafeDumper":
        exp_yaml = "foo:\n  - bar\n  - baz\nqux: quux\n"
    else:
        exp_yaml = "foo:\n- bar\n- baz\nqux: quux\n"
    assert (
        salt.utils.yamldumper.dump(
            data,
            Dumper=salt.utils.yamldumper.IndentedSafeOrderedDumper,
            default_flow_style=False,
        )
        == exp_yaml
    )


def test_yaml_defaultdict_dump():
    """
    Test yaml.dump with defaultdict
    """
    data = defaultdict(list)
    data["foo"].append("bar")
    exp_yaml = "foo: [bar]\n"
    assert salt.utils.yamldumper.safe_dump(data) == exp_yaml


def test_yaml_namespaced_dict_wrapper_dump():
    """
    Test yaml.dump with NamespacedDictWrapper
    """
    data = NamespacedDictWrapper({"test": {"foo": "bar"}}, "test")
    exp_yaml = (
        "!!python/object/new:salt.utils.context.NamespacedDictWrapper\n"
        "dictitems: {foo: bar}\n"
        "state:\n"
        "  _NamespacedDictWrapper__dict:\n"
        "    test: {foo: bar}\n"
        "  pre_keys: !!python/tuple [test]\n"
    )
    assert salt.utils.yamldumper.dump(data) == exp_yaml


def test_yaml_undefined_dump():
    """
    Test yaml.safe_dump with None
    """
    data = {"foo": None}
    exp_yaml = "{foo: null}\n"
    assert salt.utils.yamldumper.safe_dump(data) == exp_yaml


def test_yaml_hashable_ordered_dict_dump():
    """
    Test yaml.dump with HashableOrderedDict
    """
    data = HashableOrderedDict([("foo", "bar"), ("baz", "qux")])
    exp_yaml = "{foo: bar, baz: qux}\n"
    assert (
        salt.utils.yamldumper.dump(data, Dumper=salt.utils.yamldumper.OrderedDumper)
        == exp_yaml
    )



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>

    Test cases for salt.modules.gnomedesktop
"""

import pytest

import salt.modules.gnomedesktop as gnomedesktop
from tests.support.mock import patch


@pytest.fixture
def configure_loader_modules():
    return {gnomedesktop: {}}


def test_ping():
    """
    Test for A test to ensure the GNOME module is loaded
    """
    assert gnomedesktop.ping()


def test_getidledelay():
    """
    Test for Return the current idle delay setting in seconds
    """
    with patch("salt.modules.gnomedesktop._GSettings") as gsettings_mock:
        with patch.object(gsettings_mock, "_get", return_value=True):
            assert gnomedesktop.getIdleDelay()


def test_setidledelay():
    """
    Test for Set the current idle delay setting in seconds
    """
    with patch("salt.modules.gnomedesktop._GSettings") as gsettings_mock:
        with patch.object(gsettings_mock, "_set", return_value=True):
            assert gnomedesktop.setIdleDelay(5)


def test_getclockformat():
    """
    Test for Return the current clock format, either 12h or 24h format.
    """
    with patch("salt.modules.gnomedesktop._GSettings") as gsettings_mock:
        with patch.object(gsettings_mock, "_get", return_value=True):
            assert gnomedesktop.getClockFormat()


def test_setclockformat():
    """
    Test for Set the clock format, either 12h or 24h format..
    """
    with patch("salt.modules.gnomedesktop._GSettings") as gsettings_mock:
        with patch.object(gsettings_mock, "_set", return_value=True):
            assert gnomedesktop.setClockFormat("12h")

        assert not gnomedesktop.setClockFormat("a")


def test_getclockshowdate():
    """
    Test for Return the current setting, if the date is shown in the clock
    """
    with patch("salt.modules.gnomedesktop._GSettings") as gsettings_mock:
        with patch.object(gsettings_mock, "_get", return_value=True):
            assert gnomedesktop.getClockShowDate()


def test_setclockshowdate():
    """
    Test for Set whether the date is visible in the clock
    """
    with patch("salt.modules.gnomedesktop._GSettings") as gsettings_mock:
        assert not gnomedesktop.setClockShowDate("kvalue")

        with patch.object(gsettings_mock, "_get", return_value=True):
            assert gnomedesktop.setClockShowDate(True)


def test_getidleactivation():
    """
    Test for Get whether the idle activation is enabled
    """
    with patch("salt.modules.gnomedesktop._GSettings") as gsettings_mock:
        with patch.object(gsettings_mock, "_get", return_value=True):
            assert gnomedesktop.getIdleActivation()


def test_setidleactivation():
    """
    Test for Set whether the idle activation is enabled
    """
    with patch("salt.modules.gnomedesktop._GSettings") as gsettings_mock:
        assert not gnomedesktop.setIdleActivation("kvalue")

        with patch.object(gsettings_mock, "_set", return_value=True):
            assert gnomedesktop.setIdleActivation(True)


def test_get():
    """
    Test for Get key in a particular GNOME schema
    """
    with patch("salt.modules.gnomedesktop._GSettings") as gsettings_mock:
        with patch.object(gsettings_mock, "_get", return_value=True):
            assert gnomedesktop.get()


def test_set_():
    """
    Test for Set key in a particular GNOME schema.
    """
    with patch("salt.modules.gnomedesktop._GSettings") as gsettings_mock:
        with patch.object(gsettings_mock, "_get", return_value=True):
            assert gnomedesktop.set_()



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>

    Test cases for salt.modules.hg
"""

import os

import pytest

import salt.states.hg as hg
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {hg: {}}


def test_latest():
    """
    Test to Make sure the repository is cloned to
    the given directory and is up to date
    """
    ret = {"changes": {}, "comment": "", "name": "salt", "result": True}
    mock = MagicMock(return_value=True)
    with patch.object(hg, "_fail", mock):
        assert hg.latest("salt")

    mock = MagicMock(side_effect=[False, True, False, False, False, False])
    with patch.object(os.path, "isdir", mock):
        mock = MagicMock(return_value=True)
        with patch.object(hg, "_handle_existing", mock):
            assert hg.latest("salt", target="c:\\salt")

        with patch.dict(hg.__opts__, {"test": True}):
            mock = MagicMock(return_value=True)
            with patch.object(hg, "_neutral_test", mock):
                assert hg.latest("salt", target="c:\\salt")

        with patch.dict(hg.__opts__, {"test": False}):
            mock = MagicMock(return_value=True)
            with patch.object(hg, "_clone_repo", mock):
                assert hg.latest("salt", target="c:\\salt") == ret


def test_latest_update_changes():
    """
    Test to make sure we don't update even if we have changes
    """
    ret = {"changes": {}, "comment": "", "name": "salt", "result": True}
    revision_mock = MagicMock(return_value="abcdef")
    pull_mock = MagicMock(return_value="Blah.")
    update_mock = MagicMock()

    with patch.dict(
        hg.__salt__,
        {
            "hg.revision": revision_mock,
            "hg.pull": pull_mock,
            "hg.update": update_mock,
        },
    ):
        mock = MagicMock(side_effect=[True, True])
        with patch.object(os.path, "isdir", mock):
            mock = MagicMock(return_value=True)
            with patch.dict(hg.__opts__, {"test": False}):
                with patch.object(hg, "_clone_repo", mock):
                    assert hg.latest("salt", target="c:\\salt", update_head=True) == ret
                    assert update_mock.called


def test_latest_no_update_changes():
    """
    Test to make sure we don't update even if we have changes
    """
    ret = {
        "changes": {},
        "comment": (
            "Update is probably required but update_head=False so we will skip"
            " updating."
        ),
        "name": "salt",
        "result": True,
    }
    revision_mock = MagicMock(return_value="abcdef")
    pull_mock = MagicMock(return_value="Blah.")
    update_mock = MagicMock()

    with patch.dict(
        hg.__salt__,
        {
            "hg.revision": revision_mock,
            "hg.pull": pull_mock,
            "hg.update": update_mock,
        },
    ):
        mock = MagicMock(side_effect=[True, True])
        with patch.object(os.path, "isdir", mock):
            mock = MagicMock(return_value=True)
            with patch.dict(hg.__opts__, {"test": False}):
                with patch.object(hg, "_clone_repo", mock):
                    assert (
                        hg.latest("salt", target="c:\\salt", update_head=False) == ret
                    )
                    assert not update_mock.called


def test_latest_no_update_no_changes():
    """
    Test to Make sure the repository is cloned to
    the given directory and is up to date
    """
    ret = {
        "changes": {},
        "comment": "No changes found and update_head=False so will skip updating.",
        "name": "salt",
        "result": True,
    }
    revision_mock = MagicMock(return_value="abcdef")
    pull_mock = MagicMock(return_value="Blah no changes found.")
    update_mock = MagicMock()

    with patch.dict(
        hg.__salt__,
        {
            "hg.revision": revision_mock,
            "hg.pull": pull_mock,
            "hg.update": update_mock,
        },
    ):
        mock = MagicMock(side_effect=[True, True])
        with patch.object(os.path, "isdir", mock):
            mock = MagicMock(return_value=True)
            with patch.dict(hg.__opts__, {"test": False}):
                with patch.object(hg, "_clone_repo", mock):
                    assert (
                        hg.latest("salt", target="c:\\salt", update_head=False) == ret
                    )
                    assert not update_mock.called



"""
tests.unit.modules.test_bigip
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Unit tests for the bigip module
"""

import logging

import pytest

import salt.modules.bigip as bigip
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


class RequestsSession:
    def __init__(self):
        self.auth = None
        self.verify = None
        self.headers = {}


@pytest.fixture
def configure_loader_modules():
    return {bigip: {}}


def test__build_session_verify_ssl():
    requests_session = RequestsSession()
    with patch(
        "salt.modules.bigip.requests.sessions.Session",
        MagicMock(return_value=requests_session),
    ):
        bigip._build_session("username", "password")

    assert requests_session.auth == ("username", "password")
    assert requests_session.verify is True



"""
Utilities to help make requests to virtualbox

The virtualbox SDK reference can be found at http://download.virtualbox.org/virtualbox/SDKRef.pdf

This code assumes vboxapi.py from VirtualBox distribution
being in PYTHONPATH, or installed system-wide
"""

import logging
import re
import time

import salt.utils.compat
import salt.utils.data
from salt.utils.timeout import wait_for

log = logging.getLogger(__name__)


HAS_LIBS = False
try:
    import vboxapi

    HAS_LIBS = True

except ImportError:
    VirtualBoxManager = None
    log.trace("Couldn't import VirtualBox API")

_virtualboxManager = None

"""
Attributes we expect to have when converting an XPCOM object to a dict
"""
XPCOM_ATTRIBUTES = {
    "IMachine": [
        "id",
        "name",
        "accessible",
        "description",
        "groups",
        "memorySize",
        "OSTypeId",
        "state",
    ],
    "INetworkAdapter": [
        "adapterType",
        "slot",
        "enabled",
        "MACAddress",
        "bridgedInterface",
        "hostOnlyInterface",
        "internalNetwork",
        "NATNetwork",
        "genericDriver",
        "cableConnected",
        "lineSpeed",
        "lineSpeed",
    ],
}

UNKNOWN_MACHINE_STATE = ("Unknown", "This state is unknown to us. Might be new?")
MACHINE_STATE_LIST = [
    ("Null", "Null value (never used by the API)"),
    (
        "PoweredOff",
        "The machine is not running and has no saved execution state; "
        "it has either never been started or been shut down successfully.",
    ),
    (
        "Saved",
        "The machine is not currently running, but the execution state of the machine"
        " has been saved to an external file when it was running, from where it can be"
        " resumed.",
    ),
    (
        "Teleported",
        "The machine was teleported to a different host (or process) and then powered"
        " off. Take care when powering it on again may corrupt resources it shares with"
        " the teleportation target (e.g. disk and network).",
    ),
    (
        "Aborted",
        "The process running the machine has terminated abnormally. This may indicate a"
        " crash of the VM process in host execution context, or the VM process has been"
        " terminated externally.",
    ),
    ("Running", "The machine is currently being executed."),
    ("Paused", "Execution of the machine has been paused."),
    (
        "Stuck",
        "Execution of the machine has reached the 'Guru Meditation' condition. This"
        " indicates a severe error in the hypervisor itself.",
    ),
    (
        "Teleporting",
        "The machine is about to be teleported to a different host or process. It is"
        " possible to pause a machine in this state, but it will go to the"
        " TeleportingPausedVM state and it will not be possible to resume it again"
        " unless the teleportation fails.",
    ),
    (
        "LiveSnapshotting",
        "A live snapshot is being taken. The machine is running normally, but some of"
        " the runtime configuration options are inaccessible. Also, if paused while in"
        " this state it will transition to OnlineSnapshotting and it will not be resume"
        " the execution until the snapshot operation has completed.",
    ),
    (
        "Starting",
        "Machine is being started after powering it on from a zero execution state.",
    ),
    (
        "Stopping",
        "Machine is being normally stopped powering it off, or after the guest OS has"
        " initiated a shutdown sequence.",
    ),
    ("Saving", "Machine is saving its execution state to a file."),
    (
        "Restoring",
        "Execution state of the machine is being restored from a file after powering it"
        " on from the saved execution state.",
    ),
    (
        "TeleportingPausedVM",
        "The machine is being teleported to another host or process, but it is not "
        "running. This is the paused variant of the Teleporting state.",
    ),
    ("TeleportingIn", "Teleporting the machine state in from another host or process."),
    (
        "FaultTolerantSyncing",
        "The machine is being synced with a fault tolerant VM running else-where.",
    ),
    (
        "DeletingSnapshotOnline",
        "Like DeletingSnapshot , but the merging of media is ongoing in the "
        "background while the machine is running.",
    ),
    (
        "DeletingSnapshotPaused",
        "Like DeletingSnapshotOnline , but the machine was paused when "
        "the merging of differencing media was started.",
    ),
    (
        "OnlineSnapshotting",
        "Like LiveSnapshotting , but the machine was paused when the merging "
        "of differencing media was started.",
    ),
    (
        "RestoringSnapshot",
        "A machine snapshot is being restored; this typically does not take long.",
    ),
    (
        "DeletingSnapshot",
        "A machine snapshot is being deleted; this can take a long time since this may"
        " require merging differencing media. This value indicates that the machine is"
        " not running while the snapshot is being deleted.",
    ),
    ("SettingUp", "Lengthy setup operation is in progress."),
    ("Snapshotting", "Taking an (offline) snapshot."),
    (
        "FirstOnline",
        "Pseudo-state: first online state (for use in relational expressions).",
    ),
    (
        "LastOnline",
        "Pseudo-state: last online state (for use in relational expressions).",
    ),
    (
        "FirstTransient",
        "Pseudo-state: first transient state (for use in relational expressions).",
    ),
    (
        "LastTransient",
        "Pseudo-state: last transient state (for use in relational expressions).",
    ),
]
MACHINE_STATES = dict(MACHINE_STATE_LIST)

"""
Dict of states {
    <number>: ( <name>, <description> )
}
"""
MACHINE_STATES_ENUM = dict(enumerate(MACHINE_STATE_LIST))


def vb_get_manager():
    """
    Creates a 'singleton' manager to communicate with a local virtualbox hypervisor.
    @return:
    @rtype: VirtualBoxManager
    """
    global _virtualboxManager
    if _virtualboxManager is None and HAS_LIBS:
        salt.utils.compat.reload(vboxapi)
        _virtualboxManager = vboxapi.VirtualBoxManager(None, None)

    return _virtualboxManager


def vb_get_box():
    """
    Needed for certain operations in the SDK e.g creating sessions
    @return:
    @rtype: IVirtualBox
    """
    vb_get_manager()

    try:
        # This works in older versions of the SDK, but does not seem to work anymore.
        vbox = _virtualboxManager.vbox
    except AttributeError:
        vbox = _virtualboxManager.getVirtualBox()

    return vbox


def vb_get_max_network_slots():
    """
    Max number of slots any machine can have
    @return:
    @rtype: number
    """
    sysprops = vb_get_box().systemProperties
    totals = [
        sysprops.getMaxNetworkAdapters(adapter_type)
        for adapter_type in [
            1,  # PIIX3 A PIIX3 (PCI IDE ISA Xcelerator) chipset.
            2,  # ICH9 A ICH9 (I/O Controller Hub) chipset
        ]
    ]
    return sum(totals)


def vb_get_network_adapters(machine_name=None, machine=None):
    """
    A valid machine_name or a machine is needed to make this work!

    @param machine_name:
    @type machine_name: str
    @param machine:
    @type machine: IMachine
    @return: INetorkAdapter's converted to dicts
    @rtype: [dict]
    """

    if machine_name:
        machine = vb_get_box().findMachine(machine_name)
    network_adapters = []

    for i in range(vb_get_max_network_slots()):
        try:
            inetwork_adapter = machine.getNetworkAdapter(i)
            network_adapter = vb_xpcom_to_attribute_dict(
                inetwork_adapter, "INetworkAdapter"
            )
            network_adapter["properties"] = inetwork_adapter.getProperties("")
            network_adapters.append(network_adapter)
        except Exception:  # pylint: disable=broad-except
            pass

    return network_adapters


def vb_wait_for_network_address(
    timeout, step=None, machine_name=None, machine=None, wait_for_pattern=None
):
    """
    Wait until a machine has a network address to return or quit after the timeout

    @param timeout: in seconds
    @type timeout: float
    @param step: How regularly we want to check for ips (in seconds)
    @type step: float
    @param machine_name:
    @type machine_name: str
    @param machine:
    @type machine: IMachine
    @type wait_for_pattern: str
    @param wait_for_pattern:
    @type machine: str
    @return:
    @rtype: list
    """
    kwargs = {
        "machine_name": machine_name,
        "machine": machine,
        "wait_for_pattern": wait_for_pattern,
    }
    return wait_for(
        vb_get_network_addresses,
        timeout=timeout,
        step=step,
        default=[],
        func_kwargs=kwargs,
    )


def _check_session_state(xp_session, expected_state="Unlocked"):
    """
    @param xp_session:
    @type xp_session: ISession from the Virtualbox API
    @param expected_state: The constant descriptor according to the docs
    @type expected_state: str
    @return:
    @rtype: bool
    """
    state_value = getattr(
        _virtualboxManager.constants, "SessionState_" + expected_state
    )
    return xp_session.state == state_value


def vb_wait_for_session_state(xp_session, state="Unlocked", timeout=10, step=None):
    """
    Waits until a session state has been reached, checking at regular intervals.

    @param xp_session:
    @type xp_session: ISession from the Virtualbox API
    @param state: The constant descriptor according to the docs
    @type state: str
    @param timeout: in seconds
    @type timeout: int | float
    @param step: Intervals at which the value is checked
    @type step: int | float
    @return: Did we reach the state?
    @rtype: bool
    """
    args = (xp_session, state)
    wait_for(
        _check_session_state, timeout=timeout, step=step, default=False, func_args=args
    )


def vb_get_network_addresses(machine_name=None, machine=None, wait_for_pattern=None):
    """
    TODO distinguish between private and public addresses

    A valid machine_name or a machine is needed to make this work!

    !!!
    Guest prerequisite: GuestAddition
    !!!

    Thanks to Shrikant Havale for the StackOverflow answer http://stackoverflow.com/a/29335390

    More information on guest properties: https://www.virtualbox.org/manual/ch04.html#guestadd-guestprops

    @param machine_name:
    @type machine_name: str
    @param machine:
    @type machine: IMachine
    @return: All the IPv4 addresses we could get
    @rtype: str[]
    """
    if machine_name:
        machine = vb_get_box().findMachine(machine_name)

    ip_addresses = []
    log.debug("checking for power on:")
    if machine.state == _virtualboxManager.constants.MachineState_Running:

        log.debug("got power on:")

        # wait on an arbitrary named property
        # for instance use a dhcp client script to set a property via VBoxControl guestproperty set dhcp_done 1
        if wait_for_pattern and not machine.getGuestPropertyValue(wait_for_pattern):
            log.debug("waiting for pattern:%s:", wait_for_pattern)
            return None

        _total_slots = machine.getGuestPropertyValue("/VirtualBox/GuestInfo/Net/Count")

        # upon dhcp the net count drops to 0 and it takes some seconds for it to be set again
        if not _total_slots:
            log.debug("waiting for net count:%s:", wait_for_pattern)
            return None

        try:
            total_slots = int(_total_slots)
            for i in range(total_slots):
                try:
                    address = machine.getGuestPropertyValue(
                        f"/VirtualBox/GuestInfo/Net/{i}/V4/IP"
                    )
                    if address:
                        ip_addresses.append(address)
                except Exception as e:  # pylint: disable=broad-except
                    log.debug(e.message)
        except ValueError as e:
            log.debug(e.message)
            return None

    log.debug("returning ip_addresses:%s:", ip_addresses)
    return ip_addresses


def vb_list_machines(**kwargs):
    """
    Which machines does the hypervisor have
    @param kwargs: Passed to vb_xpcom_to_attribute_dict to filter the attributes
    @type kwargs: dict
    @return: Untreated dicts of the machines known to the hypervisor
    @rtype: [{}]
    """
    manager = vb_get_manager()
    machines = manager.getArray(vb_get_box(), "machines")
    return [
        vb_xpcom_to_attribute_dict(machine, "IMachine", **kwargs)
        for machine in machines
    ]


def vb_create_machine(name=None):
    """
    Creates a machine on the virtualbox hypervisor

    TODO pass more params to customize machine creation
    @param name:
    @type name: str
    @return: Representation of the created machine
    @rtype: dict
    """
    vbox = vb_get_box()
    log.info("Create virtualbox machine %s ", name)
    groups = None
    os_type_id = "Other"
    new_machine = vbox.createMachine(
        None, name, groups, os_type_id, None  # Settings file  # flags
    )
    vbox.registerMachine(new_machine)
    log.info("Finished creating %s", name)
    return vb_xpcom_to_attribute_dict(new_machine, "IMachine")


def vb_clone_vm(name=None, clone_from=None, clone_mode=0, timeout=10000, **kwargs):
    """
    Tells virtualbox to create a VM by cloning from an existing one

    @param name: Name for the new VM
    @type name: str
    @param clone_from:
    @type clone_from: str
    @param timeout: maximum time in milliseconds to wait or -1 to wait indefinitely
    @type timeout: int
    @return dict of resulting VM
    """
    vbox = vb_get_box()
    log.info("Clone virtualbox machine %s from %s", name, clone_from)

    source_machine = vbox.findMachine(clone_from)

    groups = None
    os_type_id = "Other"
    new_machine = vbox.createMachine(
        None, name, groups, os_type_id, None  # Settings file  # flags
    )

    progress = source_machine.cloneTo(
        new_machine, clone_mode, None  # CloneMode  # CloneOptions : None = Full?
    )

    progress.waitForCompletion(timeout)
    log.info("Finished cloning %s from %s", name, clone_from)

    vbox.registerMachine(new_machine)

    return vb_xpcom_to_attribute_dict(new_machine, "IMachine")


def _start_machine(machine, session):
    """
    Helper to try and start machines

    @param machine:
    @type machine: IMachine
    @param session:
    @type session: ISession
    @return:
    @rtype: IProgress or None
    """
    try:
        return machine.launchVMProcess(session, "", "")
    except Exception as e:  # pylint: disable=broad-except
        log.debug(e.message, exc_info=True)
        return None


def vb_start_vm(name=None, timeout=10000, **kwargs):
    """
    Tells Virtualbox to start up a VM.
    Blocking function!

    @param name:
    @type name: str
    @param timeout: Maximum time in milliseconds to wait or -1 to wait indefinitely
    @type timeout: int
    @return untreated dict of started VM
    """
    # Time tracking
    start_time = time.time()
    timeout_in_seconds = timeout / 1000
    max_time = start_time + timeout_in_seconds

    vbox = vb_get_box()
    machine = vbox.findMachine(name)
    session = _virtualboxManager.getSessionObject(vbox)

    log.info(
        "Starting machine %s in state %s", name, vb_machinestate_to_str(machine.state)
    )
    try:
        # Keep trying to start a machine
        args = (machine, session)
        progress = wait_for(_start_machine, timeout=timeout_in_seconds, func_args=args)
        if not progress:
            progress = machine.launchVMProcess(session, "", "")

        # We already waited for stuff, don't push it
        time_left = max_time - time.time()
        progress.waitForCompletion(time_left * 1000)
    finally:
        _virtualboxManager.closeMachineSession(session)

    # The session state should best be unlocked otherwise subsequent calls might cause problems
    time_left = max_time - time.time()
    vb_wait_for_session_state(session, timeout=time_left)
    log.info("Started machine %s", name)

    return vb_xpcom_to_attribute_dict(machine, "IMachine")


def vb_stop_vm(name=None, timeout=10000, **kwargs):
    """
    Tells Virtualbox to stop a VM.
    This is a blocking function!

    @param name:
    @type name: str
    @param timeout: Maximum time in milliseconds to wait or -1 to wait indefinitely
    @type timeout: int
    @return untreated dict of stopped VM
    """
    vbox = vb_get_box()
    machine = vbox.findMachine(name)
    log.info("Stopping machine %s", name)
    session = _virtualboxManager.openMachineSession(machine)
    try:
        console = session.console
        progress = console.powerDown()
        progress.waitForCompletion(timeout)
    finally:
        _virtualboxManager.closeMachineSession(session)
        vb_wait_for_session_state(session)
    log.info(
        "Stopped machine %s is now %s", name, vb_machinestate_to_str(machine.state)
    )
    return vb_xpcom_to_attribute_dict(machine, "IMachine")


def vb_destroy_machine(name=None, timeout=10000):
    """
    Attempts to get rid of a machine and all its files from the hypervisor
    @param name:
    @type name: str
    @param timeout int timeout in milliseconds
    """
    vbox = vb_get_box()
    log.info("Destroying machine %s", name)
    machine = vbox.findMachine(name)
    files = machine.unregister(2)
    progress = machine.deleteConfig(files)
    progress.waitForCompletion(timeout)
    log.info("Finished destroying machine %s", name)


def vb_xpcom_to_attribute_dict(
    xpcom,
    interface_name=None,
    attributes=None,
    excluded_attributes=None,
    extra_attributes=None,
):
    """
    Attempts to build a dict from an XPCOM object.
    Attributes that don't exist in the object return an empty string.

    attribute_list = list of str or tuple(str,<a class>)

    e.g attributes=[('bad_attribute', list)] --> { 'bad_attribute': [] }

    @param xpcom:
    @type xpcom:
    @param interface_name: Which interface we will be converting from.
                           Without this it's best to specify the list of attributes you want
    @type interface_name: str
    @param attributes: Overrides the attributes used from XPCOM_ATTRIBUTES
    @type attributes: attribute_list
    @param excluded_attributes: Which should be excluded in the returned dict.
                                !!These take precedence over extra_attributes!!
    @type excluded_attributes: attribute_list
    @param extra_attributes: Which should be retrieved in addition those already being retrieved
    @type extra_attributes: attribute_list
    @return:
    @rtype: dict
    """
    # Check the interface
    if interface_name:
        m = re.search(rf"XPCOM.+implementing {interface_name}", str(xpcom))
        if not m:
            # TODO maybe raise error here?
            log.warning(
                "Interface %s is unknown and cannot be converted to dict",
                interface_name,
            )
            return dict()

    interface_attributes = set(attributes or XPCOM_ATTRIBUTES.get(interface_name, []))
    if extra_attributes:
        interface_attributes = interface_attributes.union(extra_attributes)
    if excluded_attributes:
        interface_attributes = interface_attributes.difference(excluded_attributes)

    attribute_tuples = []
    for attribute in interface_attributes:
        if isinstance(attribute, tuple):
            attribute_name = attribute[0]
            attribute_class = attribute[1]
            value = (attribute_name, getattr(xpcom, attribute_name, attribute_class()))
        else:
            value = (attribute, getattr(xpcom, attribute, ""))
        attribute_tuples.append(value)

    return dict(attribute_tuples)


def treat_machine_dict(machine):
    """
    Make machine presentable for outside world.

    !!!Modifies the input machine!!!

    @param machine:
    @type machine: dict
    @return: the modified input machine
    @rtype: dict
    """
    machine.update(
        {
            "id": machine.get("id", ""),
            "image": machine.get("image", ""),
            "size": "{} MB".format(machine.get("memorySize", 0)),
            "state": machine_get_machinestate_str(machine),
            "private_ips": [],
            "public_ips": [],
        }
    )

    # Replaced keys
    if "memorySize" in machine:
        del machine["memorySize"]
    return machine


def vb_machinestate_to_str(machinestate):
    """
    Put a name to the state

    @param machinestate: from the machine state enum from XPCOM
    @type machinestate: int
    @return:
    @rtype: str
    """

    return vb_machinestate_to_tuple(machinestate)[0]


def vb_machinestate_to_description(machinestate):
    """
    Describe the given state

    @param machinestate: from the machine state enum from XPCOM
    @type machinestate: int | str
    @return:
    @rtype: str
    """
    return vb_machinestate_to_tuple(machinestate)[1]


def vb_machinestate_to_tuple(machinestate):
    """

    @param machinestate:
    @type machinestate: int | str
    @return:
    @rtype: tuple(<name>, <description>)
    """
    if isinstance(machinestate, int):
        ret = MACHINE_STATES_ENUM.get(machinestate, UNKNOWN_MACHINE_STATE)
    elif isinstance(machinestate, str):
        ret = MACHINE_STATES.get(machinestate, UNKNOWN_MACHINE_STATE)
    else:
        ret = UNKNOWN_MACHINE_STATE
    return salt.utils.data.decode(ret, preserve_tuples=True)


def machine_get_machinestate_tuple(machinedict):
    return vb_machinestate_to_tuple(machinedict.get("state"))


def machine_get_machinestate_str(machinedict):
    return vb_machinestate_to_str(machinedict.get("state"))


def vb_machine_exists(name):
    """
    Checks in with the hypervisor to see if the machine with the given name is known
    @param name:
    @type name:
    @return:
    @rtype:
    """
    try:
        vbox = vb_get_box()
        vbox.findMachine(name)
        return True
    except Exception as e:  # pylint: disable=broad-except
        if isinstance(e.message, str):
            message = e.message
        elif hasattr(e, "msg") and isinstance(getattr(e, "msg"), str):
            message = getattr(e, "msg")
        else:
            message = ""
        if 0 > message.find("Could not find a registered machine named"):
            log.error(message)

        return False


def vb_get_machine(name, **kwargs):
    """
    Attempts to fetch a machine from Virtualbox and convert it to a dict

    @param name: The unique name of the machine
    @type name:
    @param kwargs: To be passed to vb_xpcom_to_attribute_dict
    @type kwargs:
    @return:
    @rtype: dict
    """
    vbox = vb_get_box()
    machine = vbox.findMachine(name)
    return vb_xpcom_to_attribute_dict(machine, "IMachine", **kwargs)



"""
SQLite sdb Module

:maintainer:    SaltStack
:maturity:      New
:platform:      all

This module allows access to sqlite3 using an ``sdb://`` URI

Like all sdb modules, the sqlite3 module requires a configuration profile to
be configured in either the minion or master configuration file. This profile
requires very little. For example:

.. code-block:: yaml

    mysqlite:
      driver: sqlite3
      database: /tmp/sdb.sqlite
      table: sdb
      create_table: True

The ``driver`` refers to the sqlite3 module, ``database`` refers to the sqlite3
database file. ``table`` is the table within the db that will hold keys and
values (defaults to ``sdb``). The database and table will be created if they
do not exist.

Advanced Usage:
===============

Instead of a table name, it is possible to provide custom SQL statements to
create the table(s) and get and set values.

.. code-block:: yaml

    myadvanced
      driver: sqlite3
      database: /tmp/sdb-advanced.sqlite
    create_statements:
      - "CREATE TABLE advanced (a text, b text, c blob, d blob)"
      - "CREATE INDEX myidx ON advanced (a)"
    get_query: "SELECT d FROM advanced WHERE a=:key"
    set_query: "INSERT OR REPLACE INTO advanced (a, d) VALUES (:key, :value)"
"""

import codecs
import logging

import salt.utils.msgpack

try:
    import sqlite3

    HAS_SQLITE3 = True
except ImportError:
    HAS_SQLITE3 = False


DEFAULT_TABLE = "sdb"

log = logging.getLogger(__name__)

__func_alias__ = {"set_": "set"}


def __virtual__():
    """
    Only load if sqlite3 is available.
    """
    if not HAS_SQLITE3:
        return False
    return True


def _quote(s, errors="strict"):
    encodable = s.encode("utf-8", errors).decode("utf-8")

    nul_index = encodable.find("\x00")

    if nul_index >= 0:
        error = UnicodeEncodeError(
            "NUL-terminated utf-8",
            encodable,
            nul_index,
            nul_index + 1,
            "NUL not allowed",
        )
        error_handler = codecs.lookup_error(errors)
        replacement, _ = error_handler(error)
        encodable = encodable.replace("\x00", replacement)

    return '"' + encodable.replace('"', '""') + '"'


def _connect(profile):
    db = profile["database"]
    table = None
    conn = sqlite3.connect(db)
    cur = conn.cursor()
    stmts = profile.get("create_statements")
    table = profile.get("table", DEFAULT_TABLE)
    idx = _quote(table + "_idx")
    table = _quote(table)

    try:
        if stmts:
            for sql in stmts:
                cur.execute(sql)
        elif profile.get("create_table", True):
            cur.execute(f"CREATE TABLE {table} (key text, value blob)")
            cur.execute(f"CREATE UNIQUE INDEX {idx} ON {table} (key)")
    except sqlite3.OperationalError:
        pass

    return (conn, cur, table)


def set_(key, value, profile=None):
    """
    Set a key/value pair in sqlite3
    """
    if not profile:
        return False
    conn, cur, table = _connect(profile)
    value = memoryview(salt.utils.msgpack.packb(value))
    q = profile.get(
        "set_query",
        f"INSERT OR REPLACE INTO {table} VALUES (:key, :value)",
    )
    conn.execute(q, {"key": key, "value": value})
    conn.commit()
    return True


def get(key, profile=None):
    """
    Get a value from sqlite3
    """
    if not profile:
        return None
    _, cur, table = _connect(profile)
    q = profile.get("get_query", f"SELECT value FROM {table} WHERE key=:key")
    res = cur.execute(q, {"key": key})
    res = res.fetchone()
    if not res:
        return None
    return salt.utils.msgpack.unpackb(res[0], raw=False)
